from .tmcc2_command import TMCC2Command
from ..command_req import CommandReq
from ..constants import DEFAULT_BAUDRATE, DEFAULT_PORT
from .tmcc2_constants import LEGACY_EXTENDED_BLOCK_COMMAND_PREFIX, TMCC2RouteCommandEnum


class RouteCmd(TMCC2Command):
    def __init__(
        self, route: int, baudrate: int = DEFAULT_BAUDRATE, port: str = DEFAULT_PORT, server: str = None
    ) -> None:
        if route < 1 or route > 99:
            raise ValueError("Route must be between 1 and 99")
        req = CommandReq(TMCC2RouteCommandEnum.FIRE, route)
        super().__init__(TMCC2RouteCommandEnum.FIRE, req, route, 0, None, baudrate, port, server)
        self._command = self._build_command()

    def _command_prefix(self) -> bytes:
        return LEGACY_EXTENDED_BLOCK_COMMAND_PREFIX.to_bytes(1, byteorder="big")



from .tmcc1_command import TMCC1Command
from ..command_req import CommandReq
from ..constants import DEFAULT_BAUDRATE, DEFAULT_PORT, CommandScope
from ..tmcc1.tmcc1_constants import TMCC1EngineCommandEnum


class EngineCmd(TMCC1Command):
    def __init__(
        self,
        engine: int,
        command: TMCC1EngineCommandEnum,
        data: int = 0,
        scope: CommandScope = CommandScope.ENGINE,
        baudrate: int = DEFAULT_BAUDRATE,
        port: str = DEFAULT_PORT,
        server: str = None,
    ) -> None:
        if scope == CommandScope.ENGINE:
            if engine < 1 or engine > 99:
                raise ValueError("Engine must be between 1 and 99")
        elif scope == CommandScope.TRAIN:
            if engine < 1 or engine > 10:
                raise ValueError("Train must be between 1 and 10")
        req = CommandReq(command, engine, data=data, scope=scope)
        super().__init__(command, req, engine, data, scope, baudrate, port, server)
        self._command = self._build_command()



#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#





#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
import argparse
from argparse import ArgumentError, ArgumentParser, HelpFormatter
from threading import Lock
from typing import List, cast


class PyTrainArgumentParser(ArgumentParser):
    def __init__(self, *args, **kwargs) -> None:
        self._parent = kwargs.pop("parent", None)
        self._exit_on_error: bool = True
        self._error_message: str | None = None
        self._lock = Lock()
        if "parents" in kwargs:
            parents = kwargs["parents"].copy()
        else:
            parents = []
        super().__init__(*args, **kwargs)
        for parent in parents:
            parent._parent = self

    @property
    def parent(self) -> ArgumentParser:
        return self._parent

    @property
    def is_exit_on_error(self) -> bool:
        if self._exit_on_error is True and self.parent and isinstance(self.parent, PyTrainArgumentParser):
            return cast(PyTrainArgumentParser, self.parent).is_exit_on_error
        return self._exit_on_error

    def error(self, message: str) -> None:
        self._error_message = message
        if self.is_exit_on_error is True:
            super().error(message)
        else:
            raise ArgumentError(None, message)

    def exit(self, status: int = 0, message: str = None) -> None:
        self._error_message = message
        if self.is_exit_on_error is True:
            super().exit(status, message)
        else:
            raise ArgumentError(None, message)

    @property
    def error_message(self) -> str | None:
        return self._error_message

    # noinspection PyArgumentList
    def validate_args(self, args=None):
        msg = None
        with self._lock:
            eoe = self._exit_on_error
            try:
                self._exit_on_error = False
                args = self.parse_args(args)
            except ArgumentError as e:
                msg = e.message
            except Exception as e:
                msg = str(e)
            finally:
                if eoe is not None:
                    self._exit_on_error = eoe
            return args, msg

    def clear_exit_on_error(self) -> None:
        with self._lock:
            self._exit_on_error = False

    def reset_exit_on_error(self) -> None:
        with self._lock:
            self._exit_on_error = True

    # noinspection PyProtectedMember
    def remove_args(self, args: List[str]) -> None:
        for arg in args:
            for action in self._actions:
                opts = action.option_strings
                if (opts and opts[0] == arg) or action.dest == arg:
                    self._remove_action(action)
                    break

            for action in self._action_groups:
                for group_action in action._group_actions:
                    opts = group_action.option_strings
                    if (opts and opts[0] == arg) or group_action.dest == arg:
                        action._group_actions.remove(group_action)
                        break


class StripPrefixesHelpFormatter(HelpFormatter):
    """
    For help within PyTrain, we need to strip the "-" characters
    off of the names of the arguments, as the user doesn't need to
    enter them. ArgParse requires arguments to be hyphenated. We
    just want to hide this implementation detail from the user.
    """

    def add_usage(self, usage, actions, groups, prefix=None):
        for action in actions:
            opt_strs = []
            for option in action.option_strings:
                opt_strs.append(str(option).replace("-", ""))
            action.option_strings = opt_strs
        return super(StripPrefixesHelpFormatter, self).add_usage(usage, actions, groups, prefix)


# Custom argparse type representing a bounded int
class IntRange:
    def __init__(self, imin: int = None, imax: int = None):
        self.imin = imin
        self.imax = imax

    def __call__(self, arg):
        try:
            value = int(arg)
        except ValueError:
            raise self.exception()
        if (self.imin is not None and value < self.imin) or (self.imax is not None and value > self.imax):
            raise self.exception()
        return value

    def exception(self):
        if self.imin is not None and self.imax is not None:
            return argparse.ArgumentTypeError(f"Must be an integer in the range [{self.imin} - {self.imax}]")
        elif self.imin is not None:
            return argparse.ArgumentTypeError(f"Must be an integer >= {self.imin}")
        elif self.imax is not None:
            return argparse.ArgumentTypeError(f"Must be an integer <= {self.imax}")
        else:
            return argparse.ArgumentTypeError("Must be an integer")



#!/usr/bin/env python3

#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

import logging
from argparse import ArgumentParser
from typing import List

from . import CliBaseTMCC
from ..protocol.multibyte.lighting_cmd import LightingCmd
from ..protocol.multibyte.multibyte_constants import TMCC2LightingControl
from ..utils.argument_parser import PyTrainArgumentParser

log = logging.getLogger(__name__)


class LightingCli(CliBaseTMCC):
    @classmethod
    def command_parser(cls):
        lighting_parser = PyTrainArgumentParser(add_help=False)
        lighting_parser.add_argument("engine", metavar="Engine/Train", type=int, help="Engine/Train to control")
        sp = lighting_parser.add_subparsers(dest="sub_command", help="Engine/train sub-commands")
        cab = sp.add_parser("cab", help="Cab lighting options")
        cab_group = cab.add_mutually_exclusive_group()
        cab_group.add_argument(
            "-auto",
            action="store_const",
            const=TMCC2LightingControl.CAB_AUTO,
            dest="option",
            default=TMCC2LightingControl.CAB_AUTO,
            help="Cab lights auto",
        )
        cab_group.add_argument(
            "-on", action="store_const", const=TMCC2LightingControl.CAB_ON, dest="option", help="Cab lights on"
        )
        cab_group.add_argument(
            "-off", action="store_const", const=TMCC2LightingControl.CAB_OFF, dest="option", help="Cab lights off"
        )
        cab_group.add_argument(
            "-toggle",
            action="store_const",
            const=TMCC2LightingControl.CAB_TOGGLE,
            dest="option",
            help="Cab lights toggle",
        )

        car = sp.add_parser("car", help="Car cabin lighting options")
        car_group = car.add_mutually_exclusive_group()
        car_group.add_argument(
            "-auto",
            action="store_const",
            const=TMCC2LightingControl.CAR_AUTO,
            dest="option",
            default=TMCC2LightingControl.CAR_AUTO,
            help="Car cabin lights auto",
        )
        car_group.add_argument(
            "-on", action="store_const", const=TMCC2LightingControl.CAR_ON, dest="option", help="Car cabin lights on"
        )
        car_group.add_argument(
            "-off", action="store_const", const=TMCC2LightingControl.CAR_OFF, dest="option", help="Car cabin lights off"
        )

        ditch = sp.add_parser("ditch", aliases=["di"], help="Ditch lighting options")
        ditch_group = ditch.add_mutually_exclusive_group()
        ditch_group.add_argument(
            "-on",
            action="store_const",
            const=TMCC2LightingControl.DITCH_ON,
            dest="option",
            default=TMCC2LightingControl.DITCH_ON,
            help="Ditch lights on",
        )
        ditch_group.add_argument(
            "-onh",
            "--on_off_w_horn",
            action="store_const",
            const=TMCC2LightingControl.DITCH_ON_PULSE_OFF_WITH_HORN,
            dest="option",
            help="Ditch lights on, pulse off w/ horn",
        )
        ditch_group.add_argument(
            "-ofh",
            "--off_on_w_horn",
            action="store_const",
            const=TMCC2LightingControl.DITCH_OFF_PULSE_ON_WITH_HORN,
            dest="option",
            help="Ditch lights off, pulse on w/ horn",
        )
        ditch_group.add_argument(
            "-off", action="store_const", const=TMCC2LightingControl.DITCH_OFF, dest="option", help="Ditch lights off"
        )

        dog = sp.add_parser("doghouse", aliases=["dog"], help="Doghouse lighting options")
        dog_group = dog.add_mutually_exclusive_group()
        dog_group.add_argument(
            "-on",
            action="store_const",
            const=TMCC2LightingControl.DOGHOUSE_ON,
            dest="option",
            default=TMCC2LightingControl.DOGHOUSE_ON,
            help="Dog house light on",
        )
        dog_group.add_argument(
            "-off",
            action="store_const",
            const=TMCC2LightingControl.DOGHOUSE_OFF,
            dest="option",
            help="Dog house light off",
        )

        loco = sp.add_parser("engine_marker", aliases=["em"], help="Engine marker lighting options")
        loco_group = loco.add_mutually_exclusive_group()
        loco_group.add_argument(
            "-on",
            action="store_const",
            const=TMCC2LightingControl.LOCO_MARKER_ON,
            dest="option",
            default=TMCC2LightingControl.LOCO_MARKER_ON,
            help="Engine marker lights on",
        )
        loco_group.add_argument(
            "-off",
            action="store_const",
            const=TMCC2LightingControl.LOCO_MARKER_OFF,
            dest="option",
            help="Engine marker lights off",
        )
        loco_group.add_argument(
            "-auto",
            action="store_const",
            const=TMCC2LightingControl.LOCO_MARKER_AUTO,
            dest="option",
            help="Engine marker lights auto",
        )

        ground = sp.add_parser("ground", aliases=["g"], help="Ground lighting options")
        ground_group = ground.add_mutually_exclusive_group()
        ground_group.add_argument(
            "-auto",
            action="store_const",
            const=TMCC2LightingControl.GROUND_AUTO,
            dest="option",
            default=TMCC2LightingControl.GROUND_AUTO,
            help="Ground lights auto",
        )
        ground_group.add_argument(
            "-on", action="store_const", const=TMCC2LightingControl.GROUND_ON, dest="option", help="Ground lights on"
        )
        ground_group.add_argument(
            "-off", action="store_const", const=TMCC2LightingControl.GROUND_OFF, dest="option", help="Ground lights off"
        )

        hazard = sp.add_parser("hazard", aliases=["h"], help="Hazard lighting options")
        hazard_group = hazard.add_mutually_exclusive_group()
        hazard_group.add_argument(
            "-auto",
            action="store_const",
            const=TMCC2LightingControl.HAZARD_AUTO,
            dest="option",
            default=TMCC2LightingControl.HAZARD_AUTO,
            help="Hazard lights auto",
        )
        hazard_group.add_argument(
            "-on", action="store_const", const=TMCC2LightingControl.HAZARD_ON, dest="option", help="Hazard lights on"
        )
        hazard_group.add_argument(
            "-off", action="store_const", const=TMCC2LightingControl.HAZARD_OFF, dest="option", help="Hazard lights off"
        )

        mars = sp.add_parser("mars", aliases=["m"], help="Mars lighting options")
        mars_group = mars.add_mutually_exclusive_group()
        mars_group.add_argument(
            "-on",
            action="store_const",
            const=TMCC2LightingControl.MARS_ON,
            dest="option",
            default=TMCC2LightingControl.MARS_ON,
            help="Mars lights on",
        )
        mars_group.add_argument(
            "-off", action="store_const", const=TMCC2LightingControl.MARS_OFF, dest="option", help="Mars lights off"
        )

        r17 = sp.add_parser("rule17", aliases=["17"], help="Rule 17 lighting options")
        r17_group = r17.add_mutually_exclusive_group()
        r17_group.add_argument(
            "-auto",
            action="store_const",
            const=TMCC2LightingControl.RULE_17_AUTO,
            dest="option",
            default=TMCC2LightingControl.RULE_17_AUTO,
            help="Rule 17 auto",
        )
        r17_group.add_argument(
            "-on", action="store_const", const=TMCC2LightingControl.RULE_17_ON, dest="option", help="Rule 17 on"
        )
        r17_group.add_argument(
            "-off", action="store_const", const=TMCC2LightingControl.RULE_17_OFF, dest="option", help="Rule 17 off"
        )

        strobe = sp.add_parser("strobe", aliases=["s"], help="Strobe lighting options")
        strobe_group = strobe.add_mutually_exclusive_group()
        strobe_group.add_argument(
            "-on",
            action="store_const",
            const=TMCC2LightingControl.STROBE_LIGHT_ON,
            dest="option",
            default=TMCC2LightingControl.STROBE_LIGHT_ON,
            help="Strobe lights on, single-flash",
        )
        strobe_group.add_argument(
            "-df",
            "--double_flash",
            action="store_const",
            const=TMCC2LightingControl.STROBE_LIGHT_ON_DOUBLE,
            dest="option",
            help="Strobe lights on, double-flash",
        )
        strobe_group.add_argument(
            "-off",
            action="store_const",
            const=TMCC2LightingControl.STROBE_LIGHT_OFF,
            dest="option",
            help="Strobe lights off",
        )

        tender = sp.add_parser("tender_marker", aliases=["tm"], help="Tender marker lighting options")
        tender_group = tender.add_mutually_exclusive_group()
        tender_group.add_argument(
            "-on",
            action="store_const",
            const=TMCC2LightingControl.TENDER_MARKER_ON,
            dest="option",
            default=TMCC2LightingControl.TENDER_MARKER_ON,
            help="Tender marker lights on",
        )
        tender_group.add_argument(
            "-off",
            action="store_const",
            const=TMCC2LightingControl.TENDER_MARKER_OFF,
            dest="option",
            help="Tender marker lights off",
        )

        work = sp.add_parser("work_light", aliases=["w"], help="Work lighting options")
        work_group = work.add_mutually_exclusive_group()
        work_group.add_argument(
            "-on",
            action="store_const",
            const=TMCC2LightingControl.WORK_ON,
            dest="option",
            default=TMCC2LightingControl.WORK_ON,
            help="Work lights on",
        )
        work_group.add_argument(
            "-off", action="store_const", const=TMCC2LightingControl.WORK_OFF, dest="option", help="Work lights off"
        )
        work_group.add_argument(
            "-auto", action="store_const", const=TMCC2LightingControl.WORK_AUTO, dest="option", help="Work lights auto"
        )

        return PyTrainArgumentParser(
            "Lighting control", parents=[lighting_parser, cls.multi_parser(), cls.train_parser(), cls.cli_parser()]
        )

    def __init__(self, arg_parser: ArgumentParser = None, cmd_line: List[str] = None, do_fire: bool = True) -> None:
        super().__init__(arg_parser, cmd_line, do_fire)
        engine: int = self._args.engine
        option = self._args.option
        try:
            scope = self._determine_scope()
            cmd = LightingCmd(
                engine,
                TMCC2LightingControl(option),
                0,
                scope,
                baudrate=self._baudrate,
                port=self._port,
                server=self._server,
            )
            if self.do_fire:
                cmd.fire(baudrate=self._baudrate, port=self._port, server=self._server)
            self._command = cmd
        except ValueError as ve:
            log.exception(ve)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
from __future__ import annotations

import atexit
import logging
from abc import ABC, ABCMeta, abstractmethod
from threading import Condition, Event, RLock, Thread, get_ident
from tkinter import TclError
from typing import Callable, Generic, TypeVar, cast

from guizero import App, Box, Combo, PushButton, Text

from ..comm.command_listener import CommandDispatcher
from ..db.accessory_state import AccessoryState
from ..db.component_state import ComponentState, RouteState, SwitchState
from ..db.component_state_store import ComponentStateStore
from ..db.state_watcher import StateWatcher
from ..gpio.gpio_handler import GpioHandler
from ..protocol.command_req import CommandReq
from ..protocol.constants import CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1AuxCommandEnum, TMCC1RouteCommandEnum, TMCC1SwitchCommandEnum
from ..utils.path_utils import find_file

log = logging.getLogger(__name__)
S = TypeVar("S", bound=ComponentState)


class StateBasedGui(Thread, Generic[S], ABC):
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(
        self,
        title: str,
        label: str = None,
        width: int = None,
        height: int = None,
        aggrigator: ComponentStateGui = None,
        enabled_bg: str = "green",
        disabled_bg: str = "black",
        enabled_text: str = "black",
        disabled_text: str = "lightgrey",
    ) -> None:
        Thread.__init__(self, daemon=True, name=f"{title} GUI")
        self._cv = Condition(RLock())
        self._ev = Event()
        if width is None or height is None:
            try:
                from tkinter import Tk

                root = Tk()
                self.width = root.winfo_screenwidth()
                self.height = root.winfo_screenheight()
                root.destroy()
            except Exception as e:
                log.exception("Error determining window size", exc_info=e)
        else:
            self.width = width
            self.height = height
        self.title = title
        self.label = label
        self._aggrigator = aggrigator

        self._enabled_bg = enabled_bg
        self._disabled_bg = disabled_bg
        self._enabled_text = enabled_text
        self._disabled_text = disabled_text
        self.left_arrow = find_file("left_arrow.jpg")
        self.right_arrow = find_file("right_arrow.jpg")
        self.app = self.by_name = self.by_number = self.box = self.btn_box = self.y_offset = None
        self.pd_button_height = self.pd_button_width = self.left_scroll_btn = self.right_scroll_btn = None
        self.aggrigator_combo = None
        self._max_name_len = 0
        self._max_button_rows = self._max_button_cols = None
        self._first_button_col = 0
        self.sort_func = None
        self._app_active = False

        # States
        self._states = dict[int, S]()
        self._state_buttons = dict[int, PushButton]()
        self._state_watchers = dict[int, StateWatcher]()

        # listen for state changes
        self._dispatcher = CommandDispatcher.get()
        self._state_store = ComponentStateStore.get()
        self._synchronized = False
        self._sync_state = self._state_store.get_state(CommandScope.SYNC, 99)
        if self._sync_state and self._sync_state.is_synchronized is True:
            self._sync_watcher = None
            self.on_sync()
        else:
            self._sync_watcher = StateWatcher(self._sync_state, self.on_sync)

        self._is_closed = False

        # Thread-aware shutdown signaling
        self._tk_thread_id: int | None = None
        self._shutdown_flag = Event()

        # Important: don't call tkinter from atexit; only signal
        atexit.register(lambda: self._shutdown_flag.set())

    def close(self) -> None:
        # Only signal shutdown here; actual tkinter destroy happens on the GUI thread
        if not self._is_closed:
            self._is_closed = True
            self._shutdown_flag.set()

    def reset(self) -> None:
        self.close()

    @property
    def destroy_complete(self) -> Event:
        return self._ev

    # noinspection PyTypeChecker
    def on_sync(self) -> None:
        if self._sync_state.is_synchronized:
            if self._sync_watcher:
                self._sync_watcher.shutdown()
                self._sync_watcher = None
            self._synchronized = True

            # get all target states; watch for state changes
            accs = self.get_target_states()
            for acc in accs:
                nl = len(acc.road_name)
                self._max_name_len = nl if nl > self._max_name_len else self._max_name_len
                self._states[acc.tmcc_id] = acc
                self._state_watchers[acc.tmcc_id] = StateWatcher(acc, self.on_state_change_action(acc))

            # start GUI
            self.start()

    # noinspection PyTypeChecker
    def update_button(self, pd: S) -> None:
        with self._cv:
            if self.is_active(pd):
                self._state_buttons[pd.tmcc_id].bg = self._enabled_bg
                self._state_buttons[pd.tmcc_id].text_color = self._enabled_text
            else:
                self._state_buttons[pd.tmcc_id].bg = self._disabled_bg
                self._state_buttons[pd.tmcc_id].text_color = self._disabled_text

    def on_state_change_action(self, pd: S) -> Callable:
        def upd():
            self.update_button(pd)

        return upd

    def run(self) -> None:
        self._ev.clear()
        self._tk_thread_id = get_ident()
        GpioHandler.cache_handler(self)
        self.app = app = App(title=self.title, width=self.width, height=self.height)
        app.full_screen = True
        app.when_closed = self.close

        # poll for shutdown requests from other threads; this runs on the GuiZero/Tk thread
        def _poll_shutdown():
            if self._shutdown_flag.is_set():
                try:
                    app.destroy()
                except TclError:
                    pass  # ignore, we're shutting down
                return None
            return None

        app.repeat(500, _poll_shutdown)

        self.box = box = Box(app, layout="grid")
        app.bg = box.bg = "white"

        _ = Text(box, text=" ", grid=[0, 0, 6, 1], size=6, height=1, bold=True)
        _ = Text(box, text="    ", grid=[1, 1], size=24)
        if self._aggrigator:
            ag_box = Box(box, grid=[2, 1, 2, 1])
            if self.label:
                _ = Text(ag_box, text=self.label, align="left", size=24, bold=True, height="fill")
            self.aggrigator_combo = Combo(
                ag_box,
                options=self._aggrigator.guis,
                selected=self.title,
                align="right",
                command=self.on_combo_change,
            )
            self.aggrigator_combo.text_size = 24
            self.aggrigator_combo.text_bold = True
        else:
            # customize label
            label = f"{self.label} {self.title}" if self.label else self.title

            _ = Text(box, text=label, grid=[2, 1, 2, 1], size=24, bold=True)
        _ = Text(box, text="    ", grid=[4, 1], size=24)
        self.by_number = PushButton(
            box,
            text="By TMCC ID",
            grid=[2, 2],
            command=self.sort_by_number,
            padx=5,
            pady=5,
        )
        self.by_name = PushButton(
            box,
            text="By Name",
            grid=[3, 2],
            width=len("By TMCC ID"),
            command=self.sort_by_name,
            padx=5,
            pady=5,
        )
        self.by_name.text_size = self.by_number.text_size = 18
        self.by_number.text_bold = True
        _ = Text(box, text=" ", grid=[0, 3, 6, 1], size=4, height=1, bold=True)
        self.app.update()

        # add scroll btns
        sort_btn_height = self.by_number.tk.winfo_height()
        self.left_scroll_btn = PushButton(
            box,
            grid=[0, 1, 1, 2],
            enabled=False,
            image=self.left_arrow,
            height=sort_btn_height * 2,
            width=sort_btn_height * 2,
            align="left",
            command=self.scroll_left,
        )
        self.right_scroll_btn = PushButton(
            box,
            grid=[5, 1, 1, 2],
            enabled=False,
            image=self.right_arrow,
            height=sort_btn_height * 2,
            width=sort_btn_height * 2,
            align="right",
            command=self.scroll_right,
        )

        self.y_offset = self.box.tk.winfo_y() + self.box.tk.winfo_height()

        # put the buttons in a separate box
        self.btn_box = Box(app, layout="grid")

        # define power district push buttons
        self.sort_by_number()

        # Display GUI and start event loop; call blocks
        try:
            app.display()
        except TclError:
            # If Tcl is already tearing down, ignore
            pass
        finally:
            # Explicitly drop references to tkinter/guizero objects on the Tk thread
            if self._aggrigator:
                for sw in self._state_watchers.values():
                    sw.shutdown()
                self._state_watchers.clear()
            self.aggrigator_combo = None
            self.left_scroll_btn = None
            self.right_scroll_btn = None
            self.by_name = None
            self.by_number = None
            self.btn_box = None
            self.box = None
            self._state_buttons.clear()
            self.app = None
            self._ev.set()

    def on_combo_change(self, option: str) -> None:
        if option == self.title:
            return  # Noop
        else:
            self._aggrigator.cycle_gui(option)

    # noinspection PyUnusedLocal
    def _reset_state_buttons(self) -> None:
        for pdb in self._state_buttons.values():
            pdb.hide()
            pdb.destroy()
        self._state_buttons.clear()

    # noinspection PyTypeChecker
    def _make_state_buttons(self, states: list[S] = None) -> None:
        with self._cv:
            if self._app_active:
                self._reset_state_buttons()
            active_cols = {self._first_button_col, self._first_button_col + 1}
            row = 4
            col = 0

            btn_h = self.pd_button_height
            btn_y = 0
            self.right_scroll_btn.disable()
            self.left_scroll_btn.disable()

            self.btn_box.visible = False
            for pd in states:
                if btn_h is not None and btn_y is not None and self.y_offset + btn_y + btn_h > self.height:
                    if self._max_button_rows is None:
                        self._max_button_rows = row - 4
                    btn_y = 0
                    row = 4
                    col += 1
                if col in active_cols:
                    self._state_buttons[pd.tmcc_id] = PushButton(
                        self.btn_box,
                        text=f"#{pd.tmcc_id} {pd.road_name}",
                        grid=[col, row],
                        width=int(self.width / 2 / 13),
                        command=self.switch_state,
                        args=[pd],
                        padx=0,
                    )
                    self._state_buttons[pd.tmcc_id].text_size = 15
                    self._state_buttons[pd.tmcc_id].bg = self._enabled_bg if self.is_active(pd) else self._disabled_bg
                    self._state_buttons[pd.tmcc_id].text_color = (
                        self._enabled_text if self.is_active(pd) else self._disabled_text
                    )

                    # recalculate height
                    self.app.update()
                    if self.pd_button_width is None:
                        self.pd_button_width = self._state_buttons[pd.tmcc_id].tk.winfo_width()
                    if self.pd_button_height is None:
                        btn_h = self.pd_button_height = self._state_buttons[pd.tmcc_id].tk.winfo_height()
                    btn_y = self._state_buttons[pd.tmcc_id].tk.winfo_y() + btn_h
                else:
                    btn_y += btn_h
                row += 1
            # logic to hide/disable/enable scroll buttons
            if col <= 1:
                self.right_scroll_btn.hide()
                self.left_scroll_btn.hide()
            else:
                if max(active_cols) < col:
                    self.right_scroll_btn.enable()
                if max(active_cols) > 1:
                    self.left_scroll_btn.enable()
            self.btn_box.visible = True

    def sort_by_number(self) -> None:
        self.by_number.text_bold = True
        self.by_name.text_bold = False

        self.sort_func = lambda x: x.tmcc_id
        states = sorted(self._states.values(), key=self.sort_func)
        self._first_button_col = 0
        self._make_state_buttons(states)

    def sort_by_name(self) -> None:
        self.by_name.text_bold = True
        self.by_number.text_bold = False

        self.sort_func = lambda x: x.road_name.lower()
        states = sorted(self._states.values(), key=self.sort_func)
        self._first_button_col = 0
        self._make_state_buttons(states)

    def scroll_left(self) -> None:
        self._first_button_col -= 1
        states = sorted(self._states.values(), key=self.sort_func)
        self._make_state_buttons(states)

    def scroll_right(self) -> None:
        self._first_button_col += 1
        states = sorted(self._states.values(), key=self.sort_func)
        self._make_state_buttons(states)

    @abstractmethod
    def get_target_states(self) -> list[S]: ...

    @abstractmethod
    def is_active(self, state: S) -> bool: ...

    @abstractmethod
    def switch_state(self, state: S) -> bool: ...


class PowerDistrictsGui(StateBasedGui):
    def __init__(
        self,
        label: str = None,
        width: int = None,
        height: int = None,
        aggrigator: ComponentStateGui = None,
    ) -> None:
        StateBasedGui.__init__(self, "Power Districts", label, width, height, aggrigator)

    def get_target_states(self) -> list[AccessoryState]:
        pds: list[AccessoryState] = []
        accs = self._state_store.get_all(CommandScope.ACC)
        for acc in accs:
            acc = cast(AccessoryState, acc)
            if acc.is_power_district is True and acc.road_name and acc.road_name.lower() != "unused":
                pds.append(acc)
        return pds

    def is_active(self, state: AccessoryState) -> bool:
        return state.is_aux_on

    def switch_state(self, pd: AccessoryState) -> None:
        with self._cv:
            if pd.is_aux_on:
                CommandReq(TMCC1AuxCommandEnum.AUX2_OPT_ONE, pd.tmcc_id).send()
            else:
                CommandReq(TMCC1AuxCommandEnum.AUX1_OPT_ONE, pd.tmcc_id).send()


class SwitchesGui(StateBasedGui):
    def __init__(
        self,
        label: str = None,
        width: int = None,
        height: int = None,
        aggrigator: ComponentStateGui = None,
    ) -> None:
        StateBasedGui.__init__(self, "Switches", label, width, height, aggrigator, disabled_bg="red")

    def get_target_states(self) -> list[SwitchState]:
        pds: list[SwitchState] = []
        accs = self._state_store.get_all(CommandScope.SWITCH)
        for acc in accs:
            acc = cast(SwitchState, acc)
            if acc.road_name and acc.road_name.lower() != "unused":
                pds.append(acc)
        return pds

    def is_active(self, state: SwitchState) -> bool:
        return state.is_thru

    def switch_state(self, pd: SwitchState) -> None:
        with self._cv:
            if pd.is_thru:
                CommandReq(TMCC1SwitchCommandEnum.OUT, pd.tmcc_id).send()
            else:
                CommandReq(TMCC1SwitchCommandEnum.THRU, pd.tmcc_id).send()


class RoutesGui(StateBasedGui):
    def __init__(
        self,
        label: str = None,
        width: int = None,
        height: int = None,
        aggrigator: ComponentStateGui = None,
    ) -> None:
        StateBasedGui.__init__(self, "Routes", label, width, height, aggrigator, disabled_bg="red")

    def get_target_states(self) -> list[RouteState]:
        pds: list[RouteState] = []
        accs = self._state_store.get_all(CommandScope.ROUTE)
        for acc in accs:
            acc = cast(RouteState, acc)
            if acc.road_name and acc.road_name.lower() != "unused":
                pds.append(acc)
        return pds

    def is_active(self, state: RouteState) -> bool:
        return state.is_active

    def switch_state(self, pd: RouteState) -> None:
        with self._cv:
            if pd.is_active:
                CommandReq(TMCC1RouteCommandEnum.FIRE, pd.tmcc_id).send()
            else:
                CommandReq(TMCC1RouteCommandEnum.FIRE, pd.tmcc_id).send()


class AccessoriesGui(StateBasedGui):
    def __init__(
        self,
        label: str = None,
        width: int = None,
        height: int = None,
        aggrigator: ComponentStateGui = None,
    ) -> None:
        StateBasedGui.__init__(self, "Accessories", label, width, height, aggrigator)

    def get_target_states(self) -> list[AccessoryState]:
        pds: list[AccessoryState] = []
        accs = self._state_store.get_all(CommandScope.ACC)
        for acc in accs:
            acc = cast(AccessoryState, acc)
            if acc.is_power_district or acc.is_sensor_track:
                continue
            if acc.road_name and acc.road_name.lower() != "unused":
                pds.append(acc)
        return pds

    def is_active(self, state: AccessoryState) -> bool:
        return state.is_aux_on

    def switch_state(self, pd: AccessoryState) -> None:
        with self._cv:
            if pd.is_aux_on:
                CommandReq(TMCC1AuxCommandEnum.AUX2_OPT_ONE, pd.tmcc_id).send()
            else:
                CommandReq(TMCC1AuxCommandEnum.AUX2_OPT_ONE, pd.tmcc_id).send()


class ComponentStateGui(Thread):
    def __init__(
        self,
        label: str = None,
        initial: str = "Power Districts",
        width: int = None,
        height: int = None,
    ) -> None:
        super().__init__(daemon=True)
        self._ev = Event()
        self._guis = {
            # "Accessories": AccessoriesGui,
            "Power Districts": PowerDistrictsGui,
            "Switches": SwitchesGui,
            "Routes": RoutesGui,
        }
        # verify requested GUI exists:
        if initial not in self._guis:
            raise ValueError(f"Invalid initial GUI: {initial}")

        self.label = label
        if width is None or height is None:
            try:
                from tkinter import Tk

                root = Tk()
                self.width = root.winfo_screenwidth()
                self.height = root.winfo_screenheight()
                root.destroy()
            except Exception as e:
                log.exception("Error determining window size", exc_info=e)
        else:
            self.width = width
            self.height = height
        self._gui = None
        self.requested_gui = initial

        self.start()

    def run(self) -> None:
        # create the initially requested gui
        self._gui = self._guis[self.requested_gui](self.label, self.width, self.height, aggrigator=self)

        # wait for user to request a different GUI
        while True:
            # Wait for request to change GUI
            self._ev.wait()
            self._ev.clear()

            # Close/destroy previous GUI
            GpioHandler.release_handler(self._gui)

            # wait for Gui to be destroyed
            self._gui.destroy_complete.wait(10)
            self._gui.join()
            # clean up state
            self._gui = None

            # create and display new gui
            self._gui = self._guis[self.requested_gui](self.label, self.width, self.height, aggrigator=self)

    def cycle_gui(self, gui: str):
        if gui in self._guis:
            self.requested_gui = gui
            self._ev.set()

    @property
    def guis(self) -> list[str]:
        return list(self._guis.keys())



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

from typing import Dict, Any

from ..protocol.constants import CommandScope, Direction
from .irda_state import IrdaState
from .engine_state import EngineState, TrainState
from .component_state import ComponentState, L, P, SCOPE_TO_STATE_MAP, SwitchState


class BlockState(ComponentState):
    """
    Maintain the state of a Block section
    """

    def __init__(self, scope: CommandScope = CommandScope.BLOCK) -> None:
        if scope != CommandScope.BLOCK:
            raise ValueError(f"Invalid scope: {scope}")
        super().__init__(scope)
        self._block_req = None
        self._block_id = None
        self._prev_block = None
        self._next_block = None
        self._occupied_by: EngineState | TrainState | None = None
        self._occupied_direction = None
        self._direction = None
        self._occupied: bool = False
        self._flags: int = 0
        self._sensor_track: IrdaState | None = None
        self._switch: SwitchState | None = None

    def __repr__(self) -> str:
        msg = f"{self.block_id:>2}" if self.block_id else "NA"
        msg += f" {self.direction.name.lower()}" if self.direction else ""
        msg += f" Occupied: {'Yes' if self.is_occupied is True else 'No '}"
        msg += f" {self.name}" if self.name else ""
        msg += f" {self.occupied_by.scope.label} {self.occupied_by.address}" if self.occupied_by else ""
        msg += (
            f" {self.occupied_direction.label}"
            if self.occupied_direction and self.occupied_direction != Direction.UNKNOWN
            else ""
        )
        return f"Block {msg}"

    def update(self, command: L | P) -> None:
        from ..pdi.block_req import BlockReq
        from .component_state_store import ComponentStateStore

        if command:
            with self._cv:
                super().update(command)
                if isinstance(command, BlockReq):
                    self._block_req = command
                    self._block_id = command.block_id
                    if command.prev_block_id:
                        self._prev_block = ComponentStateStore.get_state(CommandScope.BLOCK, command.prev_block_id)
                    else:
                        self._prev_block = None
                    if command.next_block_id:
                        self._next_block = ComponentStateStore.get_state(CommandScope.BLOCK, command.next_block_id)
                    else:
                        self._next_block = None
                    self._flags = command.flags
                    self._direction = command.direction
                    self._occupied = command.is_occupied
                    self._occupied_direction = command.motive_direction
                    if self._sensor_track is None and command.sensor_track_id:
                        self._sensor_track = ComponentStateStore.get_state(CommandScope.IRDA, command.sensor_track_id)
                    if self._switch is None and command.switch_id:
                        self._switch = ComponentStateStore.get_state(CommandScope.SWITCH, command.switch_id)
                    if command.motive_id:
                        self._occupied_by = ComponentStateStore.get_state(command.motive_scope, command.motive_id)
                    else:
                        self._occupied_by = None
                    self.changed.set()
                    self._cv.notify_all()

    @property
    def is_known(self) -> bool:
        return self._block_req is not None

    @property
    def is_tmcc(self) -> bool:
        return False

    @property
    def is_legacy(self) -> bool:
        return False

    @property
    def is_lcs(self) -> bool:
        return False

    @property
    def block_id(self) -> int:
        return self._block_id

    @property
    def flags(self) -> int:
        return self._flags

    @property
    def is_occupied(self) -> bool:
        return self._occupied

    @property
    def occupied_by(self) -> TrainState | EngineState:
        return self._occupied_by

    @property
    def occupied_direction(self) -> Direction:
        return self._occupied_direction

    @property
    def direction(self) -> Direction:
        return self._direction

    @property
    def sensor_track(self) -> IrdaState:
        return self._sensor_track

    @property
    def switch(self) -> SwitchState:
        return self._switch

    @property
    def prev_block(self) -> BlockState:
        return self._prev_block

    @property
    def next_block(self) -> BlockState:
        return self._next_block

    def as_bytes(self) -> bytes:
        from ..pdi.block_req import BlockReq

        return BlockReq(self).as_bytes

    def as_dict(self) -> Dict[str, Any]:
        if self.occupied_by:
            motive = {
                "scope": self.occupied_by.scope.name.lower(),
                "tmcc_id": self.occupied_by.address,
            }
        else:
            motive = None
        return {
            "block_id": self.block_id,
            "name": self.road_name,
            "direction": self.direction.name.lower(),
            "sensor_track": self.sensor_track.address if self.sensor_track else None,
            "switch": self.switch.address if self.switch else None,
            "previous_block_id": self.prev_block.block_id if self.prev_block else None,
            "next_block_id": self.next_block.block_id if self.next_block else None,
            "is_occupied": self.is_occupied,
            "occupied_by": motive,
        }


SCOPE_TO_STATE_MAP.update({CommandScope.BLOCK: BlockState})



from __future__ import annotations

import sys

from .multibyte_command_req import MultiByteReq
from .multibyte_constants import TMCC2_VARIABLE_INDEX, TMCC2VariableEnum, VariableCommandDef

if sys.version_info >= (3, 11):
    from typing import List, Self
elif sys.version_info >= (3, 9):
    from typing_extensions import Self

from ..constants import DEFAULT_ADDRESS, CommandScope
from ..tmcc2.tmcc2_constants import (
    LEGACY_MULTIBYTE_COMMAND_PREFIX,
    LEGACY_TRAIN_COMMAND_PREFIX,
    TMCC2_SCOPE_TO_FIRST_BYTE_MAP,
)

"""
Commands to send/receive variable data byte commands
"""


class VariableCommandReq(MultiByteReq):
    @classmethod
    def build(
        cls,
        command: TMCC2VariableEnum,
        address: int = DEFAULT_ADDRESS,
        data_bytes: int | List[int] = None,
        scope: CommandScope = None,
    ) -> Self:
        return VariableCommandReq(command, address, data_bytes, scope)

    @classmethod
    def from_bytes(cls, param: bytes, from_tmcc_rx: bool = False, is_tmcc4: bool = False) -> Self:
        if not param or len(param) < 18:
            raise ValueError(f"Variable byte command requires at least 18 bytes {param.hex(':')}")
        if (
            len(param) >= 18
            and param[3] == param[6] == param[9] == param[12] == param[15] == LEGACY_MULTIBYTE_COMMAND_PREFIX
        ):
            index = 0x00FF & int.from_bytes(param[1:3], byteorder="big")
            if index != TMCC2_VARIABLE_INDEX:
                raise ValueError(f"Invalid Variable byte command: {param.hex(':')}")
            num_data_words = int(param[5])
            if (5 + num_data_words) * 3 != len(param):
                raise ValueError(f"Command requires {(5 + num_data_words) * 3} bytes: {param.hex(':')}")
            pi = int(param[11]) << 8 | int(param[8])
            cmd_enum = TMCC2VariableEnum.by_value(pi)
            if cmd_enum:
                address = cmd_enum.value.address_from_bytes(param[1:3])
                scope = CommandScope.ENGINE
                if int(param[0]) == LEGACY_TRAIN_COMMAND_PREFIX:
                    scope = CommandScope.TRAIN
                data_bytes = []
                # harvest all the data bytes; they are the third byte of each data word
                # data starts with word 5
                for i in range(14, 15 + (num_data_words - 1) * 3, 3):
                    data_bytes.append(param[i])
                # check checksum
                if cls.checksum(param[:-1]) != param[-1].to_bytes(1, byteorder="big"):
                    raise ValueError(
                        f"Invalid Variable byte checksum: {param.hex(':')} != {cls.checksum(param[:-1]).hex()}"
                    )
                # build_req the request and return
                cmd_req = VariableCommandReq.build(cmd_enum, address, data_bytes, scope)
                if from_tmcc_rx is True:
                    cmd_req._is_tmcc_rx = True
                return cmd_req
        raise ValueError(f"Invalid Variable byte command: {param.hex(':')}")

    def __init__(
        self,
        command_def_enum: TMCC2VariableEnum,
        address: int = DEFAULT_ADDRESS,
        data_bytes: int | List[int] = None,
        scope: CommandScope = None,
    ) -> None:
        super().__init__(command_def_enum, address, 0, scope)
        if data_bytes is not None and isinstance(data_bytes, int):
            self._data_bytes = [data_bytes]
        else:
            self._data_bytes = data_bytes if data_bytes is not None else []

    def __repr__(self) -> str:
        return super().__repr__()

    @property
    def data_bytes(self) -> List[int]:
        return self._data_bytes

    @property
    def num_bytes(self) -> int:
        """
        Returns the number of bytes in this command. Commands are comprised
        of three-byte words where:
            Word 1: Command index
            Word 2: Number of data words (N)
            Word 3: LSB of destination address
            Word 4: MSB of destination address
            Words 5 - N: Data words
            Word N + 1: Checksum
        """
        return (5 + self.command.value.num_data_bytes) * 3

    # noinspection PyTypeChecker
    @property
    def as_bytes(self) -> bytes:
        cd: VariableCommandDef = self.command_def
        byte_str = bytes()
        # first word is encoded address and 0x6F byte denoting variable byte packet
        byte_str += TMCC2_SCOPE_TO_FIRST_BYTE_MAP[self.scope].to_bytes(1, byteorder="big") + self._word_1
        # Word 2: number of data bytes/words
        byte_str += self.word_prefix + len(self.data_bytes).to_bytes(1, byteorder="big")
        # words 3 & 4: command LSB and MSB
        byte_str += self.word_prefix + cd.lsb.to_bytes(1, byteorder="big")
        byte_str += self.word_prefix + cd.msb.to_bytes(1, byteorder="big")
        # now add data words
        for data_byte in self.data_bytes:
            byte_str += self.word_prefix + data_byte.to_bytes(1, byteorder="big")
        # finally, add the checksum word
        byte_str += self.word_prefix
        byte_str += self.checksum(byte_str)
        return byte_str

    @property
    def index_byte(self) -> bytes:
        return TMCC2_VARIABLE_INDEX.to_bytes(1, byteorder="big")

    @property
    def data_byte(self) -> bytes:
        return (0x00).to_bytes(1, byteorder="big")



#!/usr/bin/env python3
#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
import logging
from argparse import ArgumentParser
from typing import List, Sequence

from ..pdi.amc2_req import Amc2Req
from ..pdi.constants import Amc2Action, PdiCommand
from ..protocol.command_base import CommandBase
from ..protocol.command_req import CommandReq
from ..protocol.constants import CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1AuxCommandEnum
from ..utils.argument_parser import IntRange, PyTrainArgumentParser
from . import CliBase

log = logging.getLogger(__name__)


class Amc2Cmd(CommandBase):
    def __init__(self, tmcc_id: int, req: Amc2Req | CommandReq, server: str = None) -> None:
        if tmcc_id < 1 or tmcc_id > 99:
            raise ValueError("Amc2 ID must be between 1 and 99")
        super().__init__(None, req, tmcc_id, CommandScope.AMC2, server=server)
        self._command = self._build_command()

    def _build_command(self) -> bytes | None:
        if isinstance(self.command_req, CommandReq) or isinstance(self.command_req, Amc2Req):
            reqs = [self.command_req]
        elif isinstance(self.command_req, Sequence):
            reqs = self.command_req
        else:
            raise ValueError(f"Invalid command request: {type(self.command_req)}")
        byte_str = bytes()
        for req in reqs:
            byte_str += req.as_bytes
        return byte_str

    def _command_prefix(self) -> bytes | None:
        pass

    def _encode_address(self, command_op: int) -> bytes | None:
        pass


class Amc2Cli(CliBase):
    """
    Issue Amc2 Commands.
    """

    @classmethod
    def command_parser(cls) -> ArgumentParser:
        from . import PROGRAM_NAME

        amc2_parser = PyTrainArgumentParser(add_help=False)
        amc2_parser.add_argument("amc2", metavar="Amc2 TMCC ID", type=int, help="Amc2 to operate")

        obj_group = amc2_parser.add_argument_group(
            "Motor or Lamp",
            "Specify the motor or lamp to control",
        )
        obj = obj_group.add_mutually_exclusive_group()
        obj.add_argument(
            "-motor",
            nargs=1,
            type=int,
            choices=[1, 2],
            action="store",
            help="Motor 1 or 2",
        )
        obj.add_argument(
            "-lamp",
            nargs=1,
            type=int,
            choices=[1, 2, 3, 4],
            action="store",
            help="Lamp 1 - 4",
        )

        opts_group = amc2_parser.add_argument_group(
            "Control options",
            "Turn the motor/lamp on or off, or set the speed/brightness level",
        )
        opts = opts_group.add_mutually_exclusive_group()
        opts.add_argument(
            "-off",
            action="store_true",
            help="Turn Amc2 port off",
        )
        opts.add_argument(
            "-on",
            action="store_true",
            help="Turn Amc2 port on",
        )
        opts.add_argument(
            "-level",
            nargs=1,
            type=IntRange(0, 100),
            action="store",
            help="Set level (speed) of Amc2 port (0 - 100)",
        )
        opts.add_argument(
            "-speed",
            nargs=1,
            type=IntRange(0, 100),
            action="store",
            help="Set speed (level) of Amc2 port (0 - 100)",
        )

        amc2_parser.add_argument(
            "-server",
            action="store",
            help=f"IP Address of {PROGRAM_NAME} server, if client. Server communicates with Base 3/LCS SER2",
        )
        # fire command
        return PyTrainArgumentParser(
            "Operate specified Amc2 (1 - 99)",
            parents=[amc2_parser],
        )

    def __init__(self, arg_parser: ArgumentParser = None, cmd_line: List[str] = None, do_fire: bool = True) -> None:
        super().__init__(arg_parser, cmd_line, do_fire)
        self._amc2 = self._args.amc2
        self._motor = self._args.motor[0] - 1 if self._args.motor else None
        self._lamp = self._args.lamp[0] - 1 if self._args.lamp else None
        self._mag = self._args.level[0] if self._args.level else self._args.speed[0] if self._args.speed else None

        if self._motor is not None:
            if self._args.on or self._args.off:
                req1 = CommandReq(TMCC1AuxCommandEnum.NUMERIC, self._amc2, data=self._motor + 1)
                if self._args.on:
                    req2 = CommandReq(TMCC1AuxCommandEnum.AUX1_OPT_ONE, self._amc2)
                else:
                    req2 = CommandReq(TMCC1AuxCommandEnum.AUX2_OPT_ONE, self._amc2)
                req = [req1, req2]
            else:
                req = Amc2Req(
                    self._amc2,
                    PdiCommand.AMC2_SET,
                    Amc2Action.MOTOR,
                    motor=self._motor,
                    speed=self._mag,
                )
        elif self._lamp is not None:
            if self._args.on or self._args.off:
                req1 = CommandReq(TMCC1AuxCommandEnum.NUMERIC, self._amc2, data=self._lamp + 3)
                if self._args.on:
                    req2 = CommandReq(TMCC1AuxCommandEnum.AUX1_OPT_ONE, self._amc2)
                else:
                    req2 = CommandReq(TMCC1AuxCommandEnum.AUX2_OPT_ONE, self._amc2)
                req = [req1, req2]
            else:
                req = Amc2Req(
                    self._amc2,
                    PdiCommand.AMC2_SET,
                    Amc2Action.LAMP,
                    lamp=self._lamp,
                    level=self._mag,
                )
        else:
            raise ValueError("Must specify motor or lamp")
        try:
            cmd = Amc2Cmd(self._amc2, req, server=self._server)
            if self.do_fire:
                cmd.fire(server=self._server)
            self._command = cmd
        except ValueError as ve:
            log.exception(ve)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#

from __future__ import annotations

from typing import cast

from ..constants import DEFAULT_ADDRESS, CommandScope
from ..tmcc2.tmcc2_constants import TMCC2EngineCommandEnum, tmcc2_speed_to_rpm
from .sequence_constants import SequenceCommandEnum
from .sequence_req import SequenceReq


class AbsoluteSpeedRpm(SequenceReq):
    def __init__(
        self,
        address: int = DEFAULT_ADDRESS,
        scope: CommandScope = CommandScope.ENGINE,
        data: int = 0,
    ) -> None:
        super().__init__(SequenceCommandEnum.ABSOLUTE_SPEED_RPM, address, scope)
        self._scope = scope
        self._data = data
        self._state = None
        self.add(TMCC2EngineCommandEnum.ABSOLUTE_SPEED, data=data, scope=scope)
        rpm = tmcc2_speed_to_rpm(data)
        self.add(TMCC2EngineCommandEnum.DIESEL_RPM, data=rpm, scope=scope)

    def _apply_data(self, new_data: int = None) -> int:
        from ...db.component_state_store import ComponentStateStore
        from ...db.engine_state import EngineState

        state = cast(EngineState, ComponentStateStore.get_state(self.scope, self.address, create=False))
        if state:
            new_speed = min(state.speed_max, self.data)
            self._data = new_speed
        else:
            new_speed = self.data

        for req_wrapper in self._requests:
            req = req_wrapper.request
            if req.command == TMCC2EngineCommandEnum.DIESEL_RPM:
                req.data = tmcc2_speed_to_rpm(new_speed)
            elif req.command == TMCC2EngineCommandEnum.ABSOLUTE_SPEED:
                req.data = new_speed
        return 0


SequenceCommandEnum.ABSOLUTE_SPEED_RPM.value.register_cmd_class(AbsoluteSpeedRpm)



import threading
import time
from typing import List, Tuple, Dict, Set

from gpiozero import GPIOPinInUse, PinInvalidPin, Button, Device

from .i2c import I2C
from .i2c_device import I2CDevice

IODIRA = 0x00  # Pin direction register
IODIRB = 0x01  # Pin direction register
IPOLA = 0x02
IPOLB = 0x03
GPINTENA = 0x04
GPINTENB = 0x05
DEFVALA = 0x06
DEFVALB = 0x07
INTCONA = 0x08
INTCONB = 0x09
IOCONA = 0x0A
IOCONB = 0x0B
GPPUA = 0x0C
GPPUB = 0x0D

INTFA = 0x0E
INTFB = 0x0F
INTCAPA = 0x10
INTCAPB = 0x11
GPIOA = 0x12
GPIOB = 0x13
OLATA = 0x14
OLATB = 0x15
ALL_OFFSET = [
    IODIRA,
    IODIRB,
    IPOLA,
    IPOLB,
    GPINTENA,
    GPINTENB,
    DEFVALA,
    DEFVALB,
    INTCONA,
    INTCONB,
    IOCONA,
    IOCONB,
    GPPUA,
    GPPUB,
    GPIOA,
    GPIOB,
    OLATA,
    OLATB,
]

BANK_BIT = 7
MIRROR_BIT = 6
SEQOP_BIT = 5
DISSLW_BIT = 4
HAEN_BIT = 3
ODR_BIT = 2
INTPOL_BIT = 1

GPA0 = 0
GPA1 = 1
GPA2 = 2
GPA3 = 3
GPA4 = 4
GPA5 = 5
GPA6 = 6
GPA7 = 7
GPB0 = 8
GPB1 = 9
GPB2 = 10
GPB3 = 11
GPB4 = 12
GPB5 = 13
GPB6 = 14
GPB7 = 15
ALL_GPIO = [GPA0, GPA1, GPA2, GPA3, GPA4, GPA5, GPA6, GPA7, GPB0, GPB1, GPB2, GPB3, GPB4, GPB5, GPB6, GPB7]

HIGH = 0xFF
LOW = 0x00

INPUT = 0xFF
OUTPUT = 0x00

ADDRESS_MAP = {
    0x00: "IODIRA",
    0x01: "IODIRB",
    0x02: "IPOLA",
    0x03: "IPOLB",
    0x04: "GPINTENA",
    0x05: "GPINTENB",
    0x06: "DEFVALA",
    0x07: "DEVFALB",
    0x08: "INTCONA",
    0x09: "INTCONB",
    0x0A: "IOCON",
    0x0B: "IOCON",
    0x0C: "GPPUA",
    0x0D: "GPPUB",
    0x0E: "INTFA",
    0x0F: "INTFB",
    0x10: "INTCAPA",
    0x11: "INTCAPB",
    0x12: "GPIOA",
    0x13: "GPIOB",
    0x14: "OLATA",
    0x15: "OLATB",
}
REGISTER_MAP = {value: key for key, value in ADDRESS_MAP.items()}


class Mcp23017:
    """
    Mcp23017 class to handle ICs register setup
    """

    def __init__(self, address: int = 0x23, i2c: I2C = None) -> None:
        self._lock = threading.RLock()
        self.address = address
        self.i2c = i2c if i2c else I2C()
        self.set_all_input()
        self.set_all_pull_up()
        self.set_all_interrupt_config()
        self.set_interrupt_mirror(True)
        self._int_pin = None
        self._int_btn = None
        self._clients: Dict[int, I2CDevice] = dict()

    @property
    def interrupt_pin(self) -> int:
        return self._int_pin

    @property
    def interrupt_btn(self) -> Button:
        return self._int_btn

    @property
    def polarities(self) -> int:
        ret = self.i2c.read_from(self.address, IPOLA)
        ret |= self.i2c.read_from(self.address, IPOLB) << 8
        return ret

    @polarities.setter
    def polarities(self, value: int) -> None:
        self.i2c.write_to(self.address, IPOLA, value & 0xFF)
        self.i2c.write_to(self.address, IPOLB, (value >> 8) & 0xFF)

    def set_polarity(self, gpio: int, value: int) -> None:
        pair = self.get_offset_gpio_tuple([IPOLA, IPOLB], gpio)
        self.set_bit_enabled(pair[0], pair[1], True if value is HIGH else False)

    def get_polarity(self, gpio: int) -> int:
        pair = self.get_offset_gpio_tuple([IPOLA, IPOLB], gpio)
        return HIGH if self.get_bit_enabled(pair[0], pair[1]) != 0 else LOW

    def set_all_polarity_low(self) -> None:
        """sets all GPIOs normal polarity"""
        self.i2c.write_to(self.address, IPOLA, LOW)
        self.i2c.write_to(self.address, IPOLB, LOW)

    def set_all_output(self) -> None:
        """sets all GPIOs as OUTPUT"""
        self.i2c.write_to(self.address, IODIRA, OUTPUT)
        self.i2c.write_to(self.address, IODIRB, OUTPUT)

    def set_all_input(self) -> None:
        """sets all GPIOs as INPUT"""
        self.i2c.write_to(self.address, IODIRA, INPUT)
        self.i2c.write_to(self.address, IODIRB, INPUT)

    def set_pin_mode(self, gpio, mode) -> None:
        """
        Sets the given GPIO to the given mode INPUT or OUTPUT
        :param gpio: the GPIO to set the mode to
        :param mode: one of INPUT or OUTPUT
        """
        pair = self.get_offset_gpio_tuple([IODIRA, IODIRB], gpio)
        self.set_bit_enabled(pair[0], pair[1], True if mode is INPUT else False)

    def get_pin_mode(self, gpio) -> int:
        """
        Gets the mode of the given GPIO (INPUT or OUTPUT)
        :param gpio: the GPIO to get the mode to
        """
        pair = self.get_offset_gpio_tuple([IODIRA, IODIRB], gpio)
        return self.get_bit_enabled(pair[0], pair[1])

    @property
    def pin_modes(self) -> int:
        ret = self.i2c.read_from(self.address, IODIRA)
        ret |= self.i2c.read_from(self.address, IODIRB) << 8
        return ret

    @pin_modes.setter
    def pin_modes(self, value: int) -> None:
        self.i2c.write_to(self.address, IODIRA, value & 0xFF)
        self.i2c.write_to(self.address, IODIRB, (value >> 8) & 0xFF)

    @property
    def pull_ups(self) -> int:
        ret = self.i2c.read_from(self.address, GPPUA)
        ret |= self.i2c.read_from(self.address, GPPUB) << 8
        return ret

    @pull_ups.setter
    def pull_ups(self, value: int) -> None:
        self.i2c.write_to(self.address, GPPUA, value & 0xFF)
        self.i2c.write_to(self.address, GPPUB, (value >> 8) & 0xFF)

    def set_all_pull_up(self) -> None:
        """turn on all pull-up resistors"""
        self.i2c.write_to(self.address, REGISTER_MAP["GPPUA"], HIGH)
        self.i2c.write_to(self.address, REGISTER_MAP["GPPUB"], HIGH)

    def get_all_pull_up(self) -> List[int]:
        """get all pull-up resistors state"""
        return [
            self.i2c.read_from(self.address, REGISTER_MAP["GPPUA"]),
            self.i2c.read_from(self.address, REGISTER_MAP["GPPUB"]),
        ]

    def set_pull_up(self, gpio: int, mode: int | bool = True) -> None:
        """
        Sets the given GPIO to the given mode INPUT or OUTPUT
        :param gpio: the GPIO to set the mode to
        :param mode: one of INPUT or OUTPUT
        """
        pair = self.get_offset_gpio_tuple([GPPUA, GPPUA], gpio)
        if isinstance(mode, bool):
            mode = HIGH if mode is True else LOW
        self.set_bit_enabled(pair[0], pair[1], True if mode is HIGH else False)

    def get_pull_up(self, gpio: int) -> bool:
        pair = self.get_offset_gpio_tuple([GPPUA, GPPUA], gpio)
        return self.get_bit_enabled(pair[0], pair[1]) != 0

    def digital_write(self, gpio, state) -> None:
        """
        Sets the given GPIO to the given state: HIGH or LOW
        :param gpio: the GPIO to set the state of
        :param state: one of HIGH or LOW
        """
        pair = self.get_offset_gpio_tuple([OLATA, OLATB], gpio)
        self.set_bit_enabled(pair[0], pair[1], True if state is HIGH else False)

    def digital_read(self, gpio) -> int:
        """
        Reads the current direction of the given GPIO
        :param gpio: the GPIO to read from
        :return:
        """
        pair = self.get_offset_gpio_tuple([GPIOA, GPIOB], gpio)
        with self._lock:
            bits = self.i2c.read_from(self.address, pair[0])
            return HIGH if (bits & (1 << pair[1])) > 0 else LOW

    def value(self, gpio: int) -> int:
        return 1 if self.digital_read(gpio) == HIGH else 0

    def set_value(self, gpio: int, value: int) -> None:
        if self.get_pin_mode(gpio) != OUTPUT:
            raise TypeError(f"GPIO {gpio} must be set as OUTPUT to set a value")
        self.digital_write(gpio, HIGH if value == 1 else LOW)

    def digital_read_all(self) -> List[int]:
        """
        Reads the current direction of the given GPIO
        :return:
        """
        return [self.i2c.read_from(self.address, GPIOA), self.i2c.read_from(self.address, GPIOB)]

    @property
    def values(self) -> int:
        ret = self.i2c.read_from(self.address, GPIOA)
        ret |= self.i2c.read_from(self.address, GPIOB) << 8
        return ret

    @values.setter
    def values(self, value: int) -> None:
        self.i2c.write_to(self.address, GPIOA, value & 0xFF)
        self.i2c.write_to(self.address, GPIOB, (value >> 8) & 0xFF)

    @property
    def io_control(self) -> int:
        return self.i2c.read_from(self.address, IOCONA)

    @io_control.setter
    def io_control(self, value: int) -> None:
        self.i2c.write_to(self.address, IOCONA, value)

    def get_all_interrupt_config(self) -> List:
        """
        Return interrupt comparison criteria
        """
        return [
            self.i2c.read_from(self.address, INTCONA),
            self.i2c.read_from(self.address, INTCONB),
        ]

    def set_all_interrupt_config(self, prev_val: bool = True) -> None:
        """
        Configure interrupt comparison criteria
        :param prev_val: True to compare to previous value, False to compare to def value
        """
        self.i2c.write_to(self.address, INTCONA, 0x00 if prev_val else 0xFF)
        self.i2c.write_to(self.address, INTCONB, 0x00 if prev_val else 0xFF)

    def clear_interrupts(self) -> None:
        with self._lock:
            self._clear_interrupts()

    def _clear_interrupts(self) -> None:
        self.i2c.read_from(self.address, INTCAPA)
        self.i2c.read_from(self.address, INTCAPB)

    def clear_int_a(self) -> None:
        with self._lock:
            self.i2c.read_from(self.address, INTCAPA)

    def clear_int_b(self) -> None:
        with self._lock:
            self.i2c.read_from(self.address, INTCAPB)

    def enable_interrupt(self, gpio) -> None:
        self._enable_interrupt(gpio, True)

    def disable_interrupt(self, gpio) -> None:
        self._enable_interrupt(gpio, False)

    def _enable_interrupt(self, gpio, enabled: bool = True) -> None:
        """
        Enables or disables the interrupt of a given GPIO
        :param gpio: the GPIO where the interrupt needs to be set,
        this needs to be one of GPAn or GPBn constants
        :param enabled: enable or disable the interrupt
        """
        pair = self.get_offset_gpio_tuple([GPINTENA, GPINTENB], gpio)
        self.set_bit_enabled(pair[0], pair[1], enabled)

    def is_interrupt_enabled(self, gpio) -> int:
        pair = self.get_offset_gpio_tuple([GPINTENA, GPINTENB], gpio)
        return self.get_bit_enabled(pair[0], pair[1])

    def enable_interrupts(self) -> None:
        self._set_interrupts(True)

    def disable_interrupts(self) -> None:
        self._set_interrupts(False)

    def _set_interrupts(self, enabled: bool = True) -> None:
        """
        Enables or disables the interrupt of all GPIOs
        :param enabled: enable or disable the interrupt
        """
        self.i2c.write_to(self.address, GPINTENA, 0xFF if enabled else 0x00)
        self.i2c.write_to(self.address, GPINTENB, 0xFF if enabled else 0x00)

    def get_all_interrupts(self) -> List:
        return [self.i2c.read_from(self.address, GPINTENA), self.i2c.read_from(self.address, GPINTENB)]

    @property
    def interrupts_on(self) -> int:
        ret = self.i2c.read_from(self.address, GPINTENA)
        ret |= self.i2c.read_from(self.address, GPINTENB) << 8
        return ret

    @interrupts_on.setter
    def interrupts_on(self, value: int) -> None:
        self.i2c.write_to(self.address, GPINTENA, value & 0xFF)
        self.i2c.write_to(self.address, GPINTENB, (value >> 8) & 0xFF)

    def set_interrupt_mirror(self, enable: bool = True) -> None:
        """
        Enables or disables the interrupt mirroring
        :param enable: enable or disable the interrupt mirroring
        """
        self.set_bit_enabled(IOCONA, MIRROR_BIT, enable)
        self.set_bit_enabled(IOCONB, MIRROR_BIT, enable)

    @property
    def interrupts(self) -> int:
        """
        Reads the interrupt registers.
        :return: an int representing the pin(s) that caused the interrupt
        """
        ret = self.i2c.read_from(self.address, INTFA)
        ret |= self.i2c.read_from(self.address, INTFB) << 8
        return ret

    @property
    def interrupted_pins(self) -> List[int]:
        """
        Reads the interrupt registers.
        :return: an int representing the pin(s) that caused the interrupt
        """
        ret = self.i2c.read_from(self.address, INTFA)
        ret |= self.i2c.read_from(self.address, INTFB) << 8
        return [pin for pin in range(16) if ret & (1 << pin)]

    @property
    def captures(self) -> int:
        """
        Reads the interrupt-captured register. It captures the GPIO port value at the time
        the interrupt occurred.
        Note: This method clears interrupts
        :return: an int representing the state of all GPIOs at the time of interrupt
        """
        ret = self.i2c.read_from(self.address, INTCAPA)
        ret |= self.i2c.read_from(self.address, INTCAPB) << 8
        return 0xFF & ~ret

    def read_interrupt_captures(self) -> Tuple[List[str], List[str]]:
        """
        Reads the interrupt-captured register. It captures the GPIO port value at the time
        the interrupt occurred.
        :return: a tuple of the INTCAPA and INTCAPB interrupt capture as a list of bit string
        """
        return self._get_list_of_interrupted_values_from(INTCAPA), self._get_list_of_interrupted_values_from(INTCAPB)

    def _get_list_of_interrupted_values_from(self, offset) -> List[str]:
        li = []
        interrupted = self.i2c.read_from(self.address, offset)
        bits = "{0:08b}".format(interrupted)
        for i in reversed(range(8)):
            li.append(bits[i])
        return li

    def read_interrupt_flags(self) -> Tuple[List[str], List[str]]:
        """
        Reads the interrupt flag which reflects the interrupt condition. A set bit
        indicates that the associated pin caused the interrupt.
        :return: a tuple of the INTFA and INTFB interrupt flags as list of bit string
        """
        return self._read_interrupt_flags_from(INTFA), self._read_interrupt_flags_from(INTFB)

    def _read_interrupt_flags_from(self, offset) -> List[str]:
        li = []
        interrupted = self.i2c.read_from(self.address, offset)
        bits = "{0:08b}".format(interrupted)
        for i in reversed(range(8)):
            li.append(bits[i])
        return li

    def read(self, offset) -> int:
        return self.i2c.read_from(self.address, offset)

    def write(self, offset, value) -> None:
        return self.i2c.write_to(self.address, offset, value)

    def set_bit_enabled(self, offset: int, gpio: int, enable: bool = True) -> None:
        state_before = self.i2c.read_from(self.address, offset)
        value = (state_before | self.bitmask(gpio)) if enable else (state_before & ~self.bitmask(gpio))
        self.i2c.write_to(self.address, offset, value)

    def get_bit_enabled(self, offset: int, gpio: int) -> int:
        state = self.i2c.read_from(self.address, offset)
        value = state & self.bitmask(gpio)
        return value

    @staticmethod
    def get_offset_gpio_tuple(offsets, gpio):
        if offsets[0] not in ALL_OFFSET or offsets[1] not in ALL_OFFSET:
            raise TypeError("offsets must contain a valid offset address. See description for help")
        if gpio not in ALL_GPIO:
            raise TypeError("pin must be one of GPAn or GPBn. See description for help")

        offset = offsets[0] if gpio < 8 else offsets[1]
        _gpio = gpio % 8
        return offset, _gpio

    @staticmethod
    def bitmask(gpio) -> int:
        return 1 << (gpio % 8)

    # noinspection PyProtectedMember
    def handle_interrupt(self) -> None:
        with self._lock:
            pull_ups = self.pull_ups
            interrupts = self.interrupted_pins
            state = None
            for i in interrupts:
                from ..gpio_handler import DEFAULT_BOUNCE_TIME

                bounce_time = DEFAULT_BOUNCE_TIME
                elapsed_bounce_time = 0
                # for every pin that generated an interrupt, if there is a
                # client associated with this pin, fire event
                if i in self._clients:
                    pull_up = (pull_ups & (1 << i)) != 0
                    client = self._clients[i]
                    if hasattr(client, "bounce_time") and client.bounce_time is not None:
                        bounce_time = client.bounce_time
                    # do the bounce time here; if multiple I2C buttons were pressed
                    # at the same time, this may mean we debounce more than we should,
                    # but we need to make sure each button has settled before we
                    # decide if its state has changed
                    if bounce_time > elapsed_bounce_time:
                        bounce_time -= elapsed_bounce_time
                        elapsed_bounce_time += bounce_time
                        time.sleep(bounce_time)
                    if state is None:
                        state = self.captures  # clears interrupts, re-enabling them!!
                    capture_bit = 1 if (state & (1 << i)) != 0 else 0
                    if pull_up is True:
                        active = capture_bit == 1
                    else:
                        active = capture_bit == 0
                    # this is part of debouncing; the current button state must match
                    # state at interrupt time if we are to consider it a new event
                    if active == client.is_active:
                        client._signal_event(active)
            if state is None:
                # make sure interrupts are always reset, either because
                # we processed them above or here
                self._clear_interrupts()

    def create_interrupt_handler(self, pin, interrupt_pin) -> None:
        if 0 <= pin <= 15:
            with self._lock:
                self._int_pin = interrupt_pin
                self._int_btn = Button(interrupt_pin)
                self._int_btn.when_activated = self.handle_interrupt
                self._clear_interrupts()  # public method is locked, would deadlock
        else:
            raise TypeError("pin must be one of GPAn or GPBn. See description for help")

    def close(self) -> None:
        with self._lock:
            if self._int_btn is not None:
                self._int_btn.close()
                self._int_btn = None
        self._int_pin = self._int_btn = None

    def register_client(self, pin: int, client: I2CDevice):
        if hasattr(client, "pin") and getattr(client, "pin") != pin:
            raise PinInvalidPin(f"{pin} is not a valid pin for {client}")
        self._clients[pin] = client

    def deregister_client(self, client: Device):
        if hasattr(client, "pin"):
            self._clients.pop(getattr(client, "pin"), None)


class Mcp23017Factory:
    _instance = None
    _lock = threading.RLock()

    @classmethod
    def build(
        cls,
        address: int = 0x23,
        pin: int = 0,
        interrupt_pin: int | str = None,
        client: I2CDevice = None,
    ) -> Mcp23017:
        if pin is None or pin < 0 or pin > 15:
            raise PinInvalidPin(f"{pin} is not a valid pin")
        with cls._lock:
            if cls._instance is None:
                cls._instance = Mcp23017Factory()
            mcp23017 = cls._instance._mcp23017s.get(address, None)
            if mcp23017 is None:
                mcp23017 = Mcp23017(address)
                cls._instance._mcp23017s[address] = mcp23017
                cls._instance._pins_in_use[address] = set()
            if pin not in cls._instance._pins_in_use[address]:
                cls._instance._pins_in_use[address].add(pin)
            else:
                raise GPIOPinInUse(f"Pin {pin} is already in use by Mcp23017 at address {hex(address)}")

            if interrupt_pin is not None:
                # The CQRobot Ocean board supports 2 interrupt lines, one for DGPIO ports 0-7 (A),
                # and another for DGPIO ports 8-15 (B). We have to make sure the given interrupt port
                # hasn't been assigned to a different I2C board or that it hasn't been assigned to
                # a different bank on this board
                assigned = cls._instance._interrupt_pins.get(interrupt_pin, None)
                if assigned is not None:
                    if assigned != mcp23017:
                        raise GPIOPinInUse(
                            f"Pin {interrupt_pin} is already in use by Mcp23017 at address {hex(assigned.address)}"
                        )
                    if mcp23017.interrupt_pin is not None and mcp23017.interrupt_pin != interrupt_pin:
                        raise PinInvalidPin(
                            f"Interrupt pin {mcp23017.interrupt_pin} already assigned "
                            f"to Mcp23017 at {hex(mcp23017.address)}"
                        )
                    # one more check, has the interrupt machinery been created? If so, we're done
                    if mcp23017.interrupt_pin == interrupt_pin:
                        mcp23017.register_client(pin, client)
                        return mcp23017
                cls._instance._interrupt_pins[interrupt_pin] = mcp23017
                mcp23017.create_interrupt_handler(pin, interrupt_pin)
                mcp23017.register_client(pin, client)
            return mcp23017

    # noinspection PyProtectedMember
    @classmethod
    def close(cls, mcp23017: Mcp23017, pin: int) -> None:
        with cls._lock:
            if cls._instance is None:
                return  # really this is an invalid state...
            if mcp23017.address in cls._instance._mcp23017s:
                cls._instance._pins_in_use[mcp23017.address].discard(pin)
                if len(cls._instance._pins_in_use[mcp23017.address]) == 0:
                    if mcp23017.interrupt_pin is not None:
                        cls._instance._interrupt_pins.pop(mcp23017.interrupt_pin)
                    mcp23017.close()
                    cls._instance._mcp23017s.pop(mcp23017.address)
                    cls._instance._pins_in_use.pop(mcp23017.address)

    def __init__(self) -> None:
        if self._initialized:
            return
        else:
            self._initialized = True
        self._mcp23017s: Dict[int, Mcp23017] = dict()
        self._pins_in_use: Dict[int, Set[int]] = dict()
        self._interrupt_pins: Dict[int, Mcp23017] = dict()

    def __new__(cls, *args, **kwargs):
        """
        Provides singleton functionality. We only want one instance
        of this class in a process
        """
        with cls._lock:
            if Mcp23017Factory._instance is None:
                # noinspection PyTypeChecker
                Mcp23017Factory._instance = super(Mcp23017Factory, cls).__new__(cls)
                Mcp23017Factory._instance._initialized = False
            return Mcp23017Factory._instance



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#


class Validations:
    @classmethod
    def validate_int(
        cls,
        value: int,
        min_value: int = None,
        max_value: int = None,
        label: str = None,
        allow_none: bool = False,
    ) -> int | None:
        if value is None and allow_none is True:
            return value
        if label is None:
            label = f"'{str(value)}'"
            suffix = ""
        else:
            label = f"{label}"
            suffix = f" ({value})"
        try:
            value = int(value)
        except ValueError:
            raise ValueError(f"{label} is not an integer")
        except TypeError:
            raise TypeError(f"{label} is not an integer")
        if min_value is not None and value < min_value:
            raise ValueError(f"{label} must be equal to or greater than {min_value}{suffix}")
        if max_value is not None and value > max_value:
            raise ValueError(f"{label} must be less than or equal to {max_value}{suffix}")
        return value

    @classmethod
    def validate_float(
        cls,
        value: float,
        min_value: float = None,
        max_value: float = None,
        label: str = None,
        allow_none: bool = False,
    ) -> float | None:
        if value is None and allow_none is True:
            return value
        if label is None:
            label = f"'{str(value)}'"
            suffix = ""
        else:
            label = f"{label}"
            suffix = f" ({value})"
        try:
            value = float(value)
        except ValueError:
            raise ValueError(f"{label} is not a floating point number")
        except TypeError:
            raise TypeError(f"{label} is not a floating point number")
        if min_value is not None and value < min_value:
            raise ValueError(f"{label} must be equal to or greater than {min_value}{suffix}")
        if max_value is not None and value > max_value:
            raise ValueError(f"{label} must be less than or equal to {max_value}{suffix}")
        return value



from __future__ import annotations

import abc
import logging
from abc import ABC
from collections import defaultdict
from enum import Enum, unique
from typing import List, TypeVar

from ..protocol.constants import CommandScope, Mixins
from .amc2_req import Amc2Req
from .asc2_req import Asc2Req
from .bpc2_req import Bpc2Req
from .constants import (
    Amc2Action,
    Asc2Action,
    Bpc2Action,
    CommonAction,
    D4Action,
    FriendlyMixins,
    IrdaAction,
    PdiAction,
    PdiCommand,
    Ser2Action,
    Stm2Action,
    WiFiAction,
)
from .irda_req import IrdaReq
from .pdi_req import PdiReq
from .stm2_req import Stm2Req

T = TypeVar("T", bound=PdiReq)

log = logging.getLogger(__name__)


class PdiDeviceConfig(ABC):
    __metaclass__ = abc.ABCMeta

    def __init__(self, device: PdiDevice, cmd: T) -> None:
        self._device = device
        self._tmcc_id: int = cmd.tmcc_id if cmd is not None else 0

    @property
    def device(self) -> PdiDevice:
        return self._device

    @property
    def tmcc_id(self) -> int:
        return self._tmcc_id

    @property
    @abc.abstractmethod
    def state_requests(self) -> List[T]: ...


class Acs2DeviceConfig(PdiDeviceConfig):
    def __init__(self, cmd: Asc2Req) -> None:
        super().__init__(PdiDevice.ASC2, cmd)
        self._mode = cmd.mode

    @property
    def mode(self) -> int:
        return self._mode

    @property
    def state_requests(self) -> List[T]:
        cmds = []
        if self._mode == 0:
            # Acc mode, 8 TMCC IDs
            for i in range(8):
                cmds.append(Asc2Req(self.tmcc_id + i, action=Asc2Action.CONTROL1))
        elif self._mode == 1:
            # Acc mode, 1 TMCC ID, latching
            cmds.append(Asc2Req(self.tmcc_id, action=Asc2Action.CONTROL2))
        elif self._mode == 2:
            # Switch mode, pulsed, 4 TMCC IDs
            for i in range(4):
                cmds.append(Asc2Req(self.tmcc_id + i, action=Asc2Action.CONTROL4))
        elif self._mode == 3:
            # Switch mode, latched, 4 TMCC IDs
            for i in range(4):
                cmds.append(Asc2Req(self.tmcc_id + i, action=Asc2Action.CONTROL5))
        return cmds


class Amc2DeviceConfig(PdiDeviceConfig):
    def __init__(self, cmd: Amc2Req) -> None:
        super().__init__(PdiDevice.AMC2, cmd)

    @property
    def state_requests(self) -> List[T]:
        return []


class Bpc2DeviceConfig(PdiDeviceConfig):
    def __init__(self, cmd: Bpc2Req) -> None:
        super().__init__(PdiDevice.BPC2, cmd)
        self._mode = cmd.mode

    @property
    def mode(self) -> int:
        return self._mode

    @property
    def state_requests(self) -> List[T]:
        cmds = []
        if self._mode == 0:
            # TR mode, 8 TMCC IDs
            for i in range(8):
                cmds.append(Bpc2Req(self.tmcc_id + i, action=Bpc2Action.CONTROL1))
        elif self._mode == 1:
            # TR mode, 1 TMCC ID
            cmds.append(Bpc2Req(self.tmcc_id, action=Bpc2Action.CONTROL2))
        elif self._mode == 2:
            # Acc mode, 8 TMCC IDs
            for i in range(8):
                cmds.append(Bpc2Req(self.tmcc_id + i, action=Bpc2Action.CONTROL3))
        elif self._mode == 3:
            # Acc mode, 1 TMCC ID
            cmds.append(Bpc2Req(self.tmcc_id, action=Bpc2Action.CONTROL4))
        return cmds


class Stm2DeviceConfig(PdiDeviceConfig):
    def __init__(self, cmd: Stm2Req) -> None:
        super().__init__(PdiDevice.STM2, cmd)
        self._mode = cmd.mode

    @property
    def mode(self) -> int:
        return self._mode

    @property
    def state_requests(self) -> List[T]:
        cmds = []
        if self._mode == 0:
            # 16 inputs, 16 TMCC IDs
            for i in range(16):
                cmds.append(Stm2Req(self.tmcc_id + i, action=Stm2Action.CONTROL1))
        elif self._mode == 2:
            # 8 input pairs, 8 TMCC IDs
            for i in range(8):
                cmds.append(Stm2Req(self.tmcc_id + i, action=Stm2Action.CONTROL1))
        return cmds


class IrdaDeviceConfig(PdiDeviceConfig):
    def __init__(self, cmd: IrdaReq) -> None:
        super().__init__(PdiDevice.IRDA, cmd)

    @property
    def state_requests(self) -> List[T]:
        cmds = [
            IrdaReq(self.tmcc_id, action=IrdaAction.INFO, scope=CommandScope.ACC),
            IrdaReq(self.tmcc_id, action=IrdaAction.CONFIG),
        ]
        return cmds


A = TypeVar("A", bound=PdiAction)


class DeviceWrapper:
    C = TypeVar("C", bound=PdiReq.__class__)
    E = TypeVar("E", bound=Enum)
    T = TypeVar("T", bound=PdiReq)
    DC = TypeVar("DC", bound=PdiDeviceConfig.__class__)

    def __init__(
        self,
        req_class: C,
        *commands: PdiCommand,
        enums: E = None,
        dev_class: DC = None,
        common_actions: bool = True,
    ) -> None:
        self.req_class = req_class
        self.enums = enums
        self.dev_class = dev_class
        self.commands = commands
        self.common_actions = common_actions
        self.get: PdiCommand = self._harvest_command("GET")
        self.set: PdiCommand = self._harvest_command("SET")
        self.rx: PdiCommand = self._harvest_command("RX")

    def __repr__(self) -> str:
        return f"{self.req_class.__class__.__name__}"

    def build(self, data: bytes | int, action: A = None) -> T:
        if isinstance(data, bytes):
            if self.enums is not None:
                action_byte = data[3] if len(data) > 3 else None
                if action_byte & 0x80 == 0x80:
                    error = True
                    action_byte = 0x7FF & action_byte
                else:
                    error = False
                action = self.enums.by_value(action_byte)
                return self.req_class(data, action=action, error=error)
            else:
                return self.req_class(data)
        else:
            return self.req_class(data, action=action)

    def firmware(self, tmcc_id: int) -> T | None:
        if self.get is not None:
            enum = self.enums.by_name("FIRMWARE")
            return self.req_class(tmcc_id, self.get, enum)
        return None

    def status(self, tmcc_id: int) -> T | None:
        if self.get is not None:
            enum = self.enums.by_name("STATUS")
            return self.req_class(tmcc_id, self.get, enum)
        return None

    def info(self, tmcc_id: int) -> T | None:
        if self.get is not None:
            enum = self.enums.by_name("INFO")
            return self.req_class(tmcc_id, self.get, enum)
        return None

    def config(self, tmcc_id: int) -> T | None:
        if self.get is not None:
            enum = self.enums.by_name("CONFIG")
            return self.req_class(tmcc_id, self.get, enum)
        return None

    def clear_errors(self, tmcc_id: int) -> T | None:
        if self.set is not None:
            enum = self.enums.by_name("CLEAR_ERRORS")
            return self.req_class(tmcc_id, self.set, enum)
        return None

    def reset(self, tmcc_id: int) -> T | None:
        if self.set is not None:
            enum = self.enums.by_name("RESET")
            return self.req_class(tmcc_id, self.set, enum)
        return None

    def identify(self, tmcc_id: int, ident: int = 1) -> T | None:
        if self.set is not None:
            enum = self.enums.by_name("IDENTIFY")
            return self.req_class(tmcc_id, self.set, enum, ident=ident)
        return None

    @property
    def is_common_actions(self) -> bool:
        return self.common_actions

    def _harvest_command(self, suffix: str) -> PdiCommand | None:
        suffix = suffix.strip().upper()
        for e in self.commands:
            if e.name.endswith(suffix):
                return e
        return None


D = TypeVar("D", bound=PdiDeviceConfig)


@unique
class PdiDevice(Mixins, FriendlyMixins):
    """
    All supported LCS/PDI devices should be listed here, along with their
    constructor class wrapped in a DeviceWrapper
    """

    from .asc2_req import Asc2Req
    from .base_req import BaseReq
    from .block_req import BlockReq
    from .bpc2_req import Bpc2Req
    from .d4_req import D4Req
    from .irda_req import IrdaReq
    from .lcs_req import Ser2Req
    from .pdi_req import AllReq, PingReq, TmccReq
    from .stm2_req import Stm2Req
    from .wifi_req import WiFiReq

    BASE = DeviceWrapper(BaseReq)
    D4 = DeviceWrapper(D4Req, enums=D4Action, common_actions=False)
    PING = DeviceWrapper(PingReq)
    ALL = DeviceWrapper(AllReq, PdiCommand.ALL_GET, PdiCommand.ALL_SET)
    TMCC = DeviceWrapper(TmccReq, PdiCommand.TMCC_TX, PdiCommand.TMCC_RX)
    TMCC4 = DeviceWrapper(TmccReq, PdiCommand.TMCC4_TX, PdiCommand.TMCC4_RX)
    WIFI = DeviceWrapper(WiFiReq, PdiCommand.WIFI_GET, PdiCommand.WIFI_SET, PdiCommand.WIFI_RX, enums=WiFiAction)
    SER2 = DeviceWrapper(Ser2Req, PdiCommand.SER2_GET, PdiCommand.SER2_SET, PdiCommand.SER2_RX, enums=Ser2Action)
    BLOCK = DeviceWrapper(BlockReq, PdiCommand.BLOCK_GET, PdiCommand.BLOCK_SET, PdiCommand.BLOCK_RX)
    IRDA = DeviceWrapper(
        IrdaReq,
        PdiCommand.IRDA_GET,
        PdiCommand.IRDA_SET,
        PdiCommand.IRDA_RX,
        enums=IrdaAction,
        dev_class=IrdaDeviceConfig,
    )
    ASC2 = DeviceWrapper(
        Asc2Req,
        PdiCommand.ASC2_GET,
        PdiCommand.ASC2_SET,
        PdiCommand.ASC2_RX,
        enums=Asc2Action,
        dev_class=Acs2DeviceConfig,
    )
    AMC2 = DeviceWrapper(
        Amc2Req,
        PdiCommand.AMC2_GET,
        PdiCommand.AMC2_SET,
        PdiCommand.AMC2_RX,
        enums=Amc2Action,
        dev_class=Amc2DeviceConfig,
    )
    BPC2 = DeviceWrapper(
        Bpc2Req,
        PdiCommand.BPC2_GET,
        PdiCommand.BPC2_SET,
        PdiCommand.BPC2_RX,
        enums=Bpc2Action,
        dev_class=Bpc2DeviceConfig,
    )
    STM2 = DeviceWrapper(
        Stm2Req,
        PdiCommand.STM2_GET,
        PdiCommand.STM2_SET,
        PdiCommand.STM2_RX,
        enums=Stm2Action,
        dev_class=Stm2DeviceConfig,
    )
    UPDATE = DeviceWrapper(BaseReq)

    @classmethod
    def from_pdi_command(cls, cmd: PdiCommand) -> PdiDevice:
        return cls(cmd.name.split("_")[0].upper())

    @classmethod
    def from_data(cls, data: bytes) -> PdiDevice:
        return cls(PdiCommand(data[1]).name.split("_")[0].upper())

    @property
    def can_build_device(self) -> bool:
        return self.value.dev_class is not None

    def build_req(self, data: bytes | int, action: A = None) -> T:
        return self.value.build(data, action)

    def build_device(self, cmd: T) -> D:
        return self.value.dev_class(cmd)

    def firmware(self, tmcc_id: int) -> T:
        return self.value.firmware(tmcc_id)

    def config(self, tmcc_id: int) -> T:
        return self.value.config(tmcc_id)

    def status(self, tmcc_id: int) -> T:
        return self.value.status(tmcc_id)

    def info(self, tmcc_id: int) -> T:
        return self.value.info(tmcc_id)

    def clear_errors(self, tmcc_id: int) -> T:
        """
        Build a Clear Errors request
        """
        return self.value.clear_errors(tmcc_id)

    def reset(self, tmcc_id: int) -> T:
        """
        Build a Reset request
        """
        return self.value.identify(tmcc_id)

    def identify(self, tmcc_id: int, ident: int = 1) -> T:
        """
        Build an Identify request
        """
        return self.value.identify(tmcc_id, ident)


class SystemDeviceDict(defaultdict):
    """
    Maintains a dictionary of CommandScope to ComponentStateDict
    """

    def __missing__(self, key: PdiDevice) -> PdiDeviceDict:
        """
        generate a ComponentState object for the dictionary, based on the key
        """
        if isinstance(key, PdiDevice):
            device = key
        else:
            raise KeyError(f"Invalid device: {key}")
        # create the component state dict for this key
        self[key] = PdiDeviceDict(device)
        return self[key]

    def register_pdi_device(self, cmd: PdiReq) -> List[T] | None:
        if cmd.action.bits == CommonAction.CONFIG.bits:
            tmcc_id = cmd.tmcc_id
            device = cmd.pdi_device
            if device.can_build_device is True:
                dev_config = device.build_device(cmd)
                self[device][tmcc_id] = dev_config
                return dev_config.state_requests
        return None


class PdiDeviceDict(dict):
    def __init__(self, device: PdiDevice):
        super().__init__()  # base class doesn't get a factory
        self._device = device

    @property
    def device(self) -> PdiDevice:
        return self._device



from threading import Thread, Event
from time import sleep

from .ads_1x15 import Ads1115
from ..gpio_handler import GpioHandler
from ...protocol.command_req import CommandReq
from ...protocol.constants import CommandScope, PROGRAM_NAME, DEFAULT_ADDRESS
from ...protocol.tmcc2.tmcc2_constants import TMCC2EngineCommandEnum


class AnalogHandler(Thread):
    def __init__(
        self,
        cmd: CommandReq,
        channel: int = 0,
        address: int = DEFAULT_ADDRESS,
        scope: CommandScope = CommandScope.ENGINE,
        repeat: int = 2,
        ignore=0,
        min_data: int = 0,
        max_data: int = None,
        send_all: bool = True,
        pause_for: float = 0.20,
        i2c_address: int = 0x48,
    ) -> None:
        self._cmd = cmd
        self._address = address if address and address > 0 else DEFAULT_ADDRESS
        self._scope = scope if scope and scope in {CommandScope.ENGINE, CommandScope.TRAIN} else CommandScope.ENGINE
        self._repeat = repeat if repeat >= 1 else 1
        super().__init__(daemon=True, name=f"{PROGRAM_NAME} Analog Handler {self.scope.label} {self.address}")
        self._adc = Ads1115(channel=channel, address=i2c_address, gain=Ads1115.PGA_4_096V, continuous=False)
        self._action = self._cmd.as_action(repeat=self._repeat)
        self._ev = Event()
        self._ignore = ignore if ignore is not None and ignore >= 0 else 0
        self._min_data = min_data if min_data is not None else 0
        self._max_data = max_data
        self._interp = GpioHandler.make_interpolator(max_data, min_data, 0.0, 3.3)
        self._send_all = send_all
        self._pause_for = pause_for if pause_for is not None and pause_for > 0 else 0.20
        self._last_sent = None
        self._is_running = True
        self.start()
        GpioHandler.cache_handler(self)
        if address != 99:
            self.resume()

    @property
    def address(self) -> int:
        return self._address

    @property
    def scope(self) -> CommandScope:
        return self._scope

    def update_action(self, address: int, scope: CommandScope = CommandScope.ENGINE) -> None:
        if self._is_running:
            self._address = address if address and address > 0 else DEFAULT_ADDRESS
            self._scope = scope if scope and scope in {CommandScope.ENGINE, CommandScope.TRAIN} else CommandScope.ENGINE
            self._cmd.address = address
            self._cmd.scope = scope
            self._action = self._cmd.as_action(repeat=self._repeat)
            if address != 99:
                self.resume()
            else:
                self.pause()

    def pause(self) -> None:
        self._ev.clear()

    def resume(self) -> None:
        if self.address and self.address > 0 and self._address != DEFAULT_ADDRESS:
            self._ev.set()

    @property
    def is_active(self) -> bool:
        return self._is_running

    @property
    def is_paused(self) -> bool:
        return not self._ev.is_set()

    def run(self) -> None:
        zero_cnt = 0
        while self._is_running:
            self._ev.wait()
            if self._is_running:
                value = self._adc.request()
                data = self._interp(value)
                # print(f"Raw: {value} Scaled: {data} Last: {self._last_sent} ({zero_cnt})")
                if data >= self._ignore:
                    new_value = max(data - self._ignore, self._min_data)
                    if self._send_all is True or new_value != self._last_sent or (new_value == 0 and zero_cnt < 3):
                        self._action(new_data=new_value)
                        self._last_sent = new_value
                        if self._send_all is False:
                            if new_value != 0:
                                zero_cnt = 0
                            else:
                                zero_cnt += 1
                sleep(self._pause_for)

    def reset(self) -> None:
        self._is_running = False
        self._ev.set()


class QuillingHorn(AnalogHandler):
    """
    Send TMCC2 Commands for the Quilling Horn Effect supported by Legacy engines.
    This class uses the TI Ads 1115 ADC converter connected to a 3.3V source (not 5.0V).
    """

    def __init__(
        self,
        channel: int = 0,
        address: int = DEFAULT_ADDRESS,
        scope: CommandScope = CommandScope.ENGINE,
        repeat: int = 2,
        i2c_address: int = 0x48,
    ) -> None:
        cmd = CommandReq(TMCC2EngineCommandEnum.QUILLING_HORN, address=address, scope=scope)
        super().__init__(
            cmd,
            channel=channel,
            address=address,
            scope=scope,
            repeat=repeat,
            ignore=2,
            max_data=17,
            pause_for=0.2,
            i2c_address=i2c_address,
        )


class TrainBrake(AnalogHandler):
    """
    Send TMCC2 Commands for the Train Brake supported by Legacy engines.
    This class uses the TI Ads 1115 ADC converter connected to a 3.3V source (not 5.0V).
    """

    def __init__(
        self,
        channel: int = 1,
        address: int = DEFAULT_ADDRESS,
        scope: CommandScope = CommandScope.ENGINE,
        repeat: int = 2,
        i2c_address: int = 0x48,
    ) -> None:
        cmd = CommandReq(TMCC2EngineCommandEnum.TRAIN_BRAKE, address=address, scope=scope)
        super().__init__(
            cmd,
            channel=channel,
            address=address,
            scope=scope,
            repeat=repeat,
            ignore=0,
            max_data=7,
            send_all=False,  # only send changes
            pause_for=0.1,
            i2c_address=i2c_address,
        )



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
from ..db.component_state_store import ComponentStateStore
from ..protocol.constants import CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1SwitchCommandEnum
from .gpio_device import GpioDevice, P
from .state_source import SwitchStateSource


class Switch(GpioDevice):
    def __init__(
        self,
        address: int,
        thru_pin: P,
        out_pin: P,
        thru_led_pin: P = None,
        out_led_pin: P = None,
        cathode: bool = True,
        initial_state: TMCC1SwitchCommandEnum = None,
    ) -> None:
        """
        Control a switch/turnout that responds to TMCC1 switch commands, such
        as Lionel Command/Control-equipped turnouts or turnouts connected to
        an LCS ACS 2 configured in "Switch" mode.

        Optionally, manage LEDs to reflect turnout state; through or out.
        Also supports bi-color LEDs with either common cathode or anode.
        """
        if initial_state is None:
            state = ComponentStateStore.get_state(CommandScope.SWITCH, address, create=False)
            if state:
                initial_state = state.state
            if initial_state is None:
                initial_state = TMCC1SwitchCommandEnum.THRU

        # make the CommandReqs
        thru_req, self.thru_btn, self.thru_led = self.make_button(
            thru_pin,
            TMCC1SwitchCommandEnum.THRU,
            address,
            led_pin=thru_led_pin,
            initially_on=initial_state == TMCC1SwitchCommandEnum.THRU,
            cathode=cathode,
        )
        out_req, self.out_btn, self.out_led = self.make_button(
            out_pin,
            TMCC1SwitchCommandEnum.OUT,
            address,
            led_pin=out_led_pin,
            initially_on=initial_state == TMCC1SwitchCommandEnum.OUT,
            cathode=cathode,
        )
        # bind actions to buttons
        thru_action = thru_req.as_action(repeat=2)
        out_action = out_req.as_action(repeat=2)

        self.thru_btn.when_pressed = self.with_on_action(thru_action, self.thru_led, self.out_led)
        self.out_btn.when_pressed = self.with_on_action(out_action, self.out_led, self.thru_led)

        if self.thru_led is not None and self.out_led is not None:
            self.cache_handler(SwitchStateSource(address, self.thru_led, self.out_led))



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from abc import ABC, ABCMeta, abstractmethod
from threading import Condition, Lock, RLock

from ..protocol.constants import CommandScope


class Watchable(ABC):
    __metaclass__ = ABCMeta

    _lock: Lock = RLock()
    _cv: Condition = Condition(_lock)

    @property
    def synchronizer(self) -> Condition:
        return self._cv

    @property
    def lock(self) -> Lock:
        return self._lock

    @property
    @abstractmethod
    def address(self) -> int: ...

    @property
    @abstractmethod
    def scope(self) -> CommandScope: ...



"""
Lionel PDI Command Protocol Constants
"""

from __future__ import annotations

from enum import Enum, IntEnum, unique

from ..protocol.constants import Mixins

# General constants
PDI_SOP: int = 0xD1
PDI_STF: int = 0xDE
PDI_EOP: int = 0xDF

# Keep-alive message
KEEP_ALIVE_CMD: bytes = bytes([0xD1, 0x29, 0xD7, 0xDF])

# Command Definitions
ALL_GET: int = 0x01
ALL_SET: int = 0x02

BASE_ENGINE: int = 0x20
BASE_TRAIN: int = 0x21
BASE_ACC: int = 0x22
BASE_BASE: int = 0x23
BASE_ROUTE: int = 0x24
BASE_SWITCH: int = 0x25
BASE_MEMORY: int = 0x26

TMCC_TX: int = 0x27
TMCC_RX: int = 0x28

PING: int = 0x29

UPDATE_ENGINE_SPEED: int = 0x2A
UPDATE_TRAIN_SPEED: int = 0x2B

TMCC4_TX: int = 0x2C
TMCC4_RX: int = 0x2D

D4_ENGINE: int = 0x2E
D4_TRAIN: int = 0x2F  # TODO: Not verified

IRDA_GET: int = 0x30
IRDA_SET: int = 0x31
IRDA_RX: int = 0x32

WIFI_GET: int = 0x34
WIFI_SET: int = 0x35
WIFI_RX: int = 0x36
WIFI_PING: int = 0x37

SER2_GET: int = 0x38
SER2_SET: int = 0x39
SER2_RX: int = 0x3A

ASC2_GET: int = 0x3C
ASC2_SET: int = 0x3D
ASC2_RX: int = 0x3E

BPC2_GET: int = 0x40
BPC2_SET: int = 0x41
BPC2_RX: int = 0x42

AMC2_GET: int = 0x44
AMC2_SET: int = 0x45
AMC2_RX: int = 0x46

STM2_GET: int = 0x4C
STM2_SET: int = 0x4D
STM2_RX: int = 0x4E

RF_RX = 0x72

BLE_RX = 0x76

BLOCK_GET: int = 0x90
BLOCK_SET: int = 0x91
BLOCK_RX: int = 0x92


class FriendlyMixins(Enum):
    @property
    def friendly(self) -> str:
        return self.name.title()


@unique
class PdiCommand(IntEnum, Mixins, FriendlyMixins):
    ALL_GET = ALL_GET
    ALL_SET = ALL_SET
    BASE_ENGINE = BASE_ENGINE
    BASE_TRAIN = BASE_TRAIN
    BASE_ACC = BASE_ACC
    BASE = BASE_BASE
    BASE_ROUTE = BASE_ROUTE
    BASE_SWITCH = BASE_SWITCH
    BASE_MEMORY = BASE_MEMORY
    TMCC_TX = TMCC_TX
    TMCC_RX = TMCC_RX
    TMCC4_TX = TMCC4_TX
    TMCC4_RX = TMCC4_RX
    D4_ENGINE = D4_ENGINE
    D4_TRAIN = D4_TRAIN
    PING = PING
    UPDATE_ENGINE_SPEED = UPDATE_ENGINE_SPEED
    UPDATE_TRAIN_SPEED = UPDATE_TRAIN_SPEED
    IRDA_GET = IRDA_GET
    IRDA_SET = IRDA_SET
    IRDA_RX = IRDA_RX
    WIFI_GET = WIFI_GET
    WIFI_SET = WIFI_SET
    WIFI_RX = WIFI_RX
    WIFI_PING = WIFI_PING
    ASC2_GET = ASC2_GET
    ASC2_SET = ASC2_SET
    ASC2_RX = ASC2_RX
    SER2_GET = SER2_GET
    SER2_SET = SER2_SET
    SER2_RX = SER2_RX
    BPC2_GET = BPC2_GET
    BPC2_SET = BPC2_SET
    BPC2_RX = BPC2_RX
    AMC2_GET = AMC2_GET
    AMC2_SET = AMC2_SET
    AMC2_RX = AMC2_RX
    STM2_GET = STM2_GET
    STM2_SET = STM2_SET
    STM2_RX = STM2_RX
    RF_RX = RF_RX
    BLE_RX = BLE_RX
    BLOCK_GET = BLOCK_GET
    BLOCK_SET = BLOCK_SET
    BLOCK_RX = BLOCK_RX

    @property
    def is_ping(self) -> bool:
        return self.value == PING

    @property
    def is_tmcc(self) -> bool:
        return self.value in {TMCC_TX, TMCC_RX, TMCC4_TX, TMCC4_RX}

    @property
    def is_base(self) -> bool:
        return self.value in {
            BASE_ENGINE,
            BASE_TRAIN,
            BASE_ACC,
            BASE_BASE,
            BASE_ROUTE,
            BASE_SWITCH,
            BASE_MEMORY,
            UPDATE_ENGINE_SPEED,
            UPDATE_TRAIN_SPEED,
        }

    @property
    def is_d4(self) -> bool:
        return self.value in {
            D4_ENGINE,
            D4_TRAIN,
        }

    @property
    def is_get(self) -> bool:
        return self.name.endswith("_GET")

    @property
    def is_set(self) -> bool:
        return self.name.endswith("_SET")

    @property
    def is_receive(self) -> bool:
        return self.name.endswith("_RX")

    @property
    def is_sendable(self) -> bool:
        return self.is_base or self.is_d4 or self.is_get or self.is_set

    @property
    def is_receivable(self) -> bool:
        return self.is_receive or self.is_d4

    @property
    def is_irda(self) -> bool:
        return self.value in {IRDA_GET, IRDA_SET, IRDA_RX}

    @property
    def is_wifi(self) -> bool:
        return self.value in {WIFI_GET, WIFI_SET, WIFI_RX, WIFI_PING}

    @property
    def is_asc2(self) -> bool:
        return self.value in {ASC2_GET, ASC2_SET, ASC2_RX}

    @property
    def is_amc2(self) -> bool:
        return self.value in {AMC2_GET, AMC2_SET, AMC2_RX}

    @property
    def is_stm2(self) -> bool:
        return self.value in {STM2_GET, STM2_SET, STM2_RX}

    @property
    def is_ser2(self) -> bool:
        return self.value in {SER2_GET, SER2_SET, SER2_RX}

    @property
    def is_bpc2(self) -> bool:
        return self.value in {BPC2_GET, BPC2_SET, BPC2_RX}

    @property
    def is_block(self) -> bool:
        return self.value in {BLOCK_GET, BLOCK_SET, BLOCK_RX}

    @property
    def is_lcs(self) -> bool:
        return (
            self.is_wifi or self.is_asc2 or self.is_irda or self.is_ser2 or self.is_bpc2 or self.is_stm2 or self.is_amc2
        )

    @property
    def as_bytes(self) -> bytes:
        return self.value.to_bytes(1)


class ActionDef:
    def __init__(self, bits: int, gettable: bool, settable: bool, responds: bool) -> None:
        self._bits = bits
        self._gettable = gettable
        self._settable = settable
        self._responds = responds

    @property
    def bits(self) -> int:
        return self._bits

    @property
    def is_gettable(self) -> bool:
        return self._gettable

    @property
    def is_settable(self) -> bool:
        return self._settable

    @property
    def is_responses(self) -> bool:
        return self._responds


@unique
class PdiAction(Mixins, FriendlyMixins):
    """
    Marker interface for all Pdi Action enums
    """

    def __repr__(self) -> str:
        opts = ""
        opts += "g" if self.value.is_gettable else "x"
        opts += "s" if self.value.is_settable else "x"
        opts += "r" if self.value.is_responses else "x"
        return f"{self.title} [{opts}]"

    @property
    def bits(self) -> int | None:
        if hasattr(self.value, "bits"):
            return self.value.bits
        else:
            return None

    @property
    def is_config(self) -> bool:
        return self.bits == CommonAction.CONFIG.bits

    @property
    def as_bytes(self) -> bytes:
        return self.bits.to_bytes(1, byteorder="big")


ACTION_QUERY = 0x10
ACTION_UPDATE = 0x11
ACTION_NEXT_REC = 0x12
ACTION_CLEAR = 0x13
ACTION_MAP = 0x18
ACTION_FIRST_REC = 0x19
ACTION_COUNT = 0x20


@unique
class D4Action(PdiAction):
    QUERY = ActionDef(ACTION_QUERY, True, False, True)
    UPDATE = ActionDef(ACTION_UPDATE, False, True, True)
    NEXT_REC = ActionDef(ACTION_NEXT_REC, True, False, True)
    CLEAR = ActionDef(ACTION_CLEAR, False, True, True)
    MAP = ActionDef(ACTION_MAP, True, False, True)
    FIRST_REC = ActionDef(ACTION_FIRST_REC, True, False, True)
    COUNT = ActionDef(ACTION_COUNT, True, False, True)


ACTION_FIRMWARE: int = 0x01
ACTION_STATUS: int = 0x02
ACTION_CONFIG: int = 0x03
ACTION_INFO: int = 0x04
ACTION_CLEAR_ERRORS: int = 0x05
ACTION_RESET: int = 0x06
ACTION_IDENTIFY: int = 0x07


@unique
class CommonAction(PdiAction):
    FIRMWARE = ActionDef(ACTION_FIRMWARE, True, False, True)
    STATUS = ActionDef(ACTION_STATUS, True, False, True)
    CONFIG = ActionDef(ACTION_CONFIG, True, True, True)
    INFO = ActionDef(ACTION_INFO, True, False, True)
    CLEAR_ERRORS = ActionDef(ACTION_CLEAR_ERRORS, False, True, False)
    RESET = ActionDef(ACTION_RESET, False, True, False)
    IDENTIFY = ActionDef(ACTION_IDENTIFY, False, True, False)


ACTION_WIFI_CONNECT: int = 0x10
ACTION_WIFI_IP: int = 0x11
ACTION_WIFI_RESPBCASTS: int = 0x12
ACTION_WIFI_UNLOCK: int = 0x13
ACTION_WIFI_PASSCODE: int = 0x14


@unique
class WiFiAction(PdiAction):
    FIRMWARE = ActionDef(ACTION_FIRMWARE, True, False, True)
    STATUS = ActionDef(ACTION_STATUS, True, False, True)
    CONFIG = ActionDef(ACTION_CONFIG, True, True, True)
    INFO = ActionDef(ACTION_INFO, True, False, True)
    CLEAR_ERRORS = ActionDef(ACTION_CLEAR_ERRORS, False, True, False)
    RESET = ActionDef(ACTION_RESET, False, True, False)
    IDENTIFY = ActionDef(ACTION_IDENTIFY, False, True, False)
    CONNECT = ActionDef(ACTION_WIFI_CONNECT, True, False, True)
    IP = ActionDef(ACTION_WIFI_IP, True, False, True)
    RESPBCASTS = ActionDef(ACTION_WIFI_RESPBCASTS, True, True, True)
    UNLOCK = ActionDef(ACTION_WIFI_UNLOCK, False, True, True)
    PASSCODE = ActionDef(ACTION_WIFI_PASSCODE, True, True, True)


ACTION_DATA: int = 0x10
ACTION_SEQUENCE: int = 0x11
ACTION_RECORD: int = 0x12
ACTION_DIAG_DATA: int = 0x13


@unique
class IrdaAction(PdiAction):
    FIRMWARE = ActionDef(ACTION_FIRMWARE, True, False, True)
    STATUS = ActionDef(ACTION_STATUS, True, False, True)
    CONFIG = ActionDef(ACTION_CONFIG, True, True, True)
    INFO = ActionDef(ACTION_INFO, True, False, True)
    CLEAR_ERRORS = ActionDef(ACTION_CLEAR_ERRORS, False, True, False)
    RESET = ActionDef(ACTION_RESET, False, True, False)
    IDENTIFY = ActionDef(ACTION_IDENTIFY, False, True, False)
    DATA = ActionDef(ACTION_DATA, False, False, True)
    SEQUENCE = ActionDef(ACTION_SEQUENCE, True, True, True)
    RECORD = ActionDef(ACTION_RECORD, True, True, True)
    DIAG_DATA = ActionDef(ACTION_DIAG_DATA, True, True, True)


ACTION_CONTROL1: int = 0x10
ACTION_CONTROL2: int = 0x11
ACTION_CONTROL3: int = 0x12
ACTION_CONTROL4: int = 0x14
ACTION_CONTROL5: int = 0x15


@unique
class Asc2Action(PdiAction):
    FIRMWARE = ActionDef(ACTION_FIRMWARE, True, False, True)
    STATUS = ActionDef(ACTION_STATUS, True, False, True)
    CONFIG = ActionDef(ACTION_CONFIG, True, True, True)
    INFO = ActionDef(ACTION_INFO, True, False, True)
    CLEAR_ERRORS = ActionDef(ACTION_CLEAR_ERRORS, False, True, False)
    RESET = ActionDef(ACTION_RESET, False, True, False)
    IDENTIFY = ActionDef(ACTION_IDENTIFY, False, True, False)
    CONTROL1 = ActionDef(ACTION_CONTROL1, True, True, True)
    CONTROL2 = ActionDef(ACTION_CONTROL2, True, True, True)
    CONTROL3 = ActionDef(ACTION_CONTROL3, True, True, True)
    CONTROL4 = ActionDef(ACTION_CONTROL4, True, True, True)
    CONTROL5 = ActionDef(ACTION_CONTROL5, True, True, True)


ACTION_MOTOR: int = 0x10
ACTION_LAMP: int = 0x11
ACTION_MOTOR_CONFIG: int = 0x12


@unique
class Amc2Action(PdiAction):
    FIRMWARE = ActionDef(ACTION_FIRMWARE, True, False, True)
    STATUS = ActionDef(ACTION_STATUS, True, False, True)
    CONFIG = ActionDef(ACTION_CONFIG, True, True, True)
    INFO = ActionDef(ACTION_INFO, True, False, True)
    CLEAR_ERRORS = ActionDef(ACTION_CLEAR_ERRORS, False, True, False)
    RESET = ActionDef(ACTION_RESET, False, True, False)
    IDENTIFY = ActionDef(ACTION_IDENTIFY, False, True, False)
    MOTOR = ActionDef(ACTION_MOTOR, True, True, True)
    LAMP = ActionDef(ACTION_LAMP, True, True, True)
    MOTOR_CONFIG = ActionDef(ACTION_MOTOR_CONFIG, True, True, True)


@unique
class Ser2Action(PdiAction):
    FIRMWARE = ActionDef(ACTION_FIRMWARE, True, False, True)
    STATUS = ActionDef(ACTION_STATUS, True, False, True)
    CONFIG = ActionDef(ACTION_CONFIG, True, True, True)
    INFO = ActionDef(ACTION_INFO, True, False, True)
    CLEAR_ERRORS = ActionDef(ACTION_CLEAR_ERRORS, False, True, False)
    RESET = ActionDef(ACTION_RESET, False, True, False)
    IDENTIFY = ActionDef(ACTION_IDENTIFY, False, True, False)


ACTION_CONTROL4_BPC2: int = 0x13


@unique
class Bpc2Action(PdiAction):
    FIRMWARE = ActionDef(ACTION_FIRMWARE, True, False, True)
    STATUS = ActionDef(ACTION_STATUS, True, False, True)
    CONFIG = ActionDef(ACTION_CONFIG, True, True, True)
    INFO = ActionDef(ACTION_INFO, True, False, True)
    CLEAR_ERRORS = ActionDef(ACTION_CLEAR_ERRORS, False, True, False)
    RESET = ActionDef(ACTION_RESET, False, True, False)
    IDENTIFY = ActionDef(ACTION_IDENTIFY, False, True, False)
    CONTROL1 = ActionDef(ACTION_CONTROL1, True, True, True)
    CONTROL2 = ActionDef(ACTION_CONTROL2, True, True, True)
    CONTROL3 = ActionDef(ACTION_CONTROL3, True, True, True)
    CONTROL4 = ActionDef(ACTION_CONTROL4_BPC2, True, True, True)


@unique
class Stm2Action(PdiAction):
    FIRMWARE = ActionDef(ACTION_FIRMWARE, True, False, True)
    STATUS = ActionDef(ACTION_STATUS, True, False, True)
    CONFIG = ActionDef(ACTION_CONFIG, True, True, True)
    INFO = ActionDef(ACTION_INFO, True, False, True)
    CLEAR_ERRORS = ActionDef(ACTION_CLEAR_ERRORS, False, True, False)
    RESET = ActionDef(ACTION_RESET, False, True, False)
    IDENTIFY = ActionDef(ACTION_IDENTIFY, False, True, False)
    CONTROL1 = ActionDef(ACTION_CONTROL1, True, False, True)


ALL_FIRMWARE = {
    Asc2Action.FIRMWARE,
    Amc2Action.FIRMWARE,
    Bpc2Action.FIRMWARE,
    IrdaAction.FIRMWARE,
    Ser2Action.FIRMWARE,
    Stm2Action.FIRMWARE,
    Stm2Action.FIRMWARE,
    WiFiAction.FIRMWARE,
}

ALL_STATUS = {
    Asc2Action.STATUS,
    Amc2Action.STATUS,
    Bpc2Action.STATUS,
    IrdaAction.STATUS,
    Ser2Action.STATUS,
    Stm2Action.STATUS,
    Stm2Action.STATUS,
    WiFiAction.STATUS,
}

ALL_CONFIG = {
    Asc2Action.CONFIG,
    Amc2Action.CONFIG,
    Bpc2Action.CONFIG,
    IrdaAction.CONFIG,
    Ser2Action.CONFIG,
    Stm2Action.CONFIG,
    Stm2Action.CONFIG,
    WiFiAction.CONFIG,
}

ALL_INFO = {
    Asc2Action.INFO,
    Amc2Action.INFO,
    Bpc2Action.INFO,
    IrdaAction.INFO,
    Ser2Action.INFO,
    Stm2Action.INFO,
    Stm2Action.INFO,
    WiFiAction.INFO,
}

ALL_RESET = {
    Asc2Action.RESET,
    Amc2Action.RESET,
    Bpc2Action.RESET,
    IrdaAction.RESET,
    Ser2Action.RESET,
    Stm2Action.RESET,
    Stm2Action.RESET,
    WiFiAction.RESET,
}

ALL_IDENTIFY = {
    Asc2Action.IDENTIFY,
    Amc2Action.IDENTIFY,
    Bpc2Action.IDENTIFY,
    IrdaAction.IDENTIFY,
    Ser2Action.IDENTIFY,
    Stm2Action.IDENTIFY,
    Stm2Action.IDENTIFY,
    WiFiAction.IDENTIFY,
}

ALL_RXs = {e for e in PdiCommand if e.name.endswith("_RX")}
ALL_SETs = {e for e in PdiCommand if e.name.endswith("_SET")}
ALL_GETs = {e for e in PdiCommand if e.name.endswith("_GET")}



#!/usr/bin/env python3

#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

import logging
from argparse import ArgumentParser
from typing import List

from . import CliBaseTMCC
from ..protocol.multibyte.multibyte_constants import TMCC2RailSoundsEffectsControl
from ..protocol.multibyte.sound_effects_cmd import SoundEffectsCmd
from ..utils.argument_parser import PyTrainArgumentParser

log = logging.getLogger(__name__)


class SoundEffectsCli(CliBaseTMCC):
    @classmethod
    def command_parser(cls):
        sounds_parser = PyTrainArgumentParser(add_help=False)
        sounds_parser.add_argument("engine", metavar="Engine/Train", type=int, help="Engine/Train to control")
        sounds_parser.add_argument(
            "-add_fuel",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.ADD_FUEL,
            dest="option",
            help="Increment diesel fuel load by 10 gallons",
        )
        sounds_parser.add_argument(
            "-b+",
            "-blend_up",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.BLEND_UP,
            dest="option",
            help="Blend level up",
        )
        sounds_parser.add_argument(
            "-b-",
            "-blend_down",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.BLEND_DOWN,
            dest="option",
            help="Blend level down",
        )
        sounds_parser.add_argument(
            "-reset_odometer",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.RESET_ODOMETER,
            dest="option",
            help="Reset odometer to zero (0)",
        )
        sounds_parser.add_argument(
            "-wheel_slip",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.WHEEL_SLIP,
            dest="option",
            help="Wheel slip trigger",
        )
        sounds_parser.add_argument(
            "-m+",
            "-volume_up",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.VOLUME_UP_RS,
            dest="option",
            help="Increase master volume",
        )
        sounds_parser.add_argument(
            "-m-",
            "-volume_down",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.VOLUME_DOWN_RS,
            dest="option",
            help="Decrease master volume",
        )
        # sounds_parser.add_argument("-off",
        #                            action="store_const",
        #                            const='SOUND_OFF',
        #                            dest='option',
        #                            help="Turn all sounds off")
        # sounds_parser.add_argument("-on",
        #                            action="store_const",
        #                            const='SOUND_ON',
        #                            dest='option',
        #                            help="Turn all sounds on")

        sp = sounds_parser.add_subparsers(dest="sub_command", help="Engine/train sub-commands")

        breaker = sp.add_parser("breaker", aliases=["br"], help="Circuit Breaker RailSounds options")
        breaker_group = breaker.add_mutually_exclusive_group()
        breaker_group.add_argument(
            "-main",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.MAIN_BREAKER,
            dest="option",
            default=TMCC2RailSoundsEffectsControl.MAIN_BREAKER,
            help="Circuit Breaker sound - Main Lights",
        )
        breaker_group.add_argument(
            "-cab",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.CAB_BREAKER,
            dest="option",
            help="Circuit Breaker sound - CAB Lights",
        )
        breaker_group.add_argument(
            "-work_lights",
            "-ground_lights",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.WORK_BREAKER,
            dest="option",
            help="Circuit Breaker sound - Work/Ground Lights",
        )

        coupler = sp.add_parser("coupler", aliases=["co"], help="Coupler impact RailSounds options")
        coupler_group = coupler.add_mutually_exclusive_group()
        coupler_group.add_argument(
            "-compress",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.COUPLER_COMPRESS,
            dest="option",
            help="Force Coupler Impact - compress",
        )
        coupler_group.add_argument(
            "-stretch",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.COUPLER_STRETCH,
            dest="option",
            help="Force Coupler Impact - stretch",
        )

        cyl = sp.add_parser("cylinder", aliases=["cy"], help="Cylinder cock clearing RailSounds options")
        cyl_group = cyl.add_mutually_exclusive_group()
        cyl_group.add_argument(
            "-on",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.CYLINDER_ON,
            dest="option",
            default=TMCC2RailSoundsEffectsControl.CYLINDER_ON,
            help="Cylinder clearing sound on",
        )
        cyl_group.add_argument(
            "-off",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.CYLINDER_OFF,
            dest="option",
            help="Cylinder clearing sound off",
        )

        pm = sp.add_parser("prime_mover", aliases=["pm"], help="Prime mover RailSounds options")
        pm_group = pm.add_mutually_exclusive_group()
        pm_group.add_argument(
            "-on",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.PRIME_ON,
            dest="option",
            default=TMCC2RailSoundsEffectsControl.PRIME_ON,
            help="Prime mover sound on",
        )
        pm_group.add_argument(
            "-off",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.PRIME_OFF,
            dest="option",
            help="Prime mover sound off",
        )

        seq = sp.add_parser("sequence", aliases=["se"], help="Sequence control RailSounds options")
        seq_group = seq.add_mutually_exclusive_group()
        seq_group.add_argument(
            "-on",
            "-enable",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.SEQUENCE_CONTROL_ON,
            dest="option",
            default=TMCC2RailSoundsEffectsControl.SEQUENCE_CONTROL_ON,
            help="Enable RailSounds sequence control",
        )
        seq_group.add_argument(
            "-off",
            "-disable",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.SEQUENCE_CONTROL_OFF,
            dest="option",
            help="Disable RailSounds sequence control",
        )

        standby = sp.add_parser("standby", aliases=["st"], help="Standby mode RailSounds options")
        standby_group = standby.add_mutually_exclusive_group()
        standby_group.add_argument(
            "-on",
            "-enable",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.STANDBY_ENABLE,
            dest="option",
            default=TMCC2RailSoundsEffectsControl.STANDBY_ENABLE,
            help="Enable RailSounds standby mode",
        )
        standby_group.add_argument(
            "-off",
            "-disable",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.STANDBY_DISABLE,
            dest="option",
            help="Disable RailSounds standby mode",
        )
        standby_group.add_argument(
            "-warning_bell",
            action="store_const",
            const=TMCC2RailSoundsEffectsControl.STANDBY_BELL,
            dest="option",
            help="Standby warning bell on",
        )

        return PyTrainArgumentParser(
            "RailSounds sound controls",
            parents=[sounds_parser, cls.multi_parser(), cls.train_parser(), cls.cli_parser()],
        )

    def __init__(self, arg_parser: ArgumentParser = None, cmd_line: List[str] = None, do_fire: bool = True) -> None:
        super().__init__(arg_parser, cmd_line, do_fire)
        engine: int = self._args.engine
        option = self._args.option
        try:
            scope = self._determine_scope()
            cmd = SoundEffectsCmd(
                engine,
                TMCC2RailSoundsEffectsControl(option),
                0,
                scope,
                baudrate=self._baudrate,
                port=self._port,
                server=self._server,
            )
            if self.do_fire:
                cmd.fire(baudrate=self._baudrate, port=self._port, server=self._server)
            self._command = cmd
        except ValueError as ve:
            log.exception(ve)



#!/usr/bin/env python3

#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

import logging
from argparse import ArgumentParser
from datetime import datetime
from typing import List

from . import CliBase
from ..comm.command_listener import CommandListener
from ..protocol.command_req import CommandReq

log = logging.getLogger(__name__)


class EchoCli:
    def __init__(self, arg_parser: ArgumentParser = None, cmd_line: List[str] = None) -> None:
        if arg_parser is None:
            arg_parser = ArgumentParser("Echo LCS SER2 output to console", parents=[CliBase.cli_parser()])
        if cmd_line is None:
            self._args = arg_parser.parse_args()
        else:
            self._args = arg_parser.parse_args(cmd_line)
        try:
            print(f"Echoing commands received by the LCS Ser2 on {self._args.port} (Ctrl-C to quit)")
            self._listener = CommandListener(self._args.baudrate, self._args.port)
            self._listener.subscribe_any(self)
        except KeyboardInterrupt:
            self._listener.shutdown()

    def __call__(self, cmd: CommandReq) -> None:
        log.info(f"{datetime.now().strftime('%H:%M:%S.%f')[:-3]} {cmd}")



from __future__ import annotations

from enum import unique, IntEnum

from ..protocol.tmcc1.tmcc1_constants import TMCC1SwitchCommandEnum
from ..protocol.command_req import CommandReq
from ..protocol.constants import Mixins


@unique
class UnitBits(Mixins, IntEnum):
    SINGLE = 0b0
    HEAD = 0b1
    MIDDLE = 0b10
    TAIL = 0b11


class ConsistComponent:
    @classmethod
    def from_bytes(cls, data: bytes) -> list[ConsistComponent]:
        consist_components: list[ConsistComponent] = []
        data_len = len(data)
        for i in range(0, 32, 2):
            if data_len > i:
                if data[i] != 0xFF and data[i + 1] != 0xFF:
                    consist_components.insert(0, ConsistComponent(tmcc_id=data[i + 1], flags=data[i]))
            else:
                break
        return consist_components

    @classmethod
    def to_bytes(cls, components: list[ConsistComponent]) -> bytes:
        byte_str = bytes()
        for comp in reversed(components):
            byte_str += comp.as_bytes
        byte_str += b"\xff" * (32 - len(byte_str))
        return byte_str

    def __init__(self, tmcc_id: int, flags: int) -> None:
        self.tmcc_id = tmcc_id
        self.flags = flags

    def __repr__(self) -> str:
        d = "F" if self.is_forward else "R"
        tl = " T" if self.is_train_linked else ""
        hm = " H" if self.is_horn_masked else ""
        dm = " D" if self.is_dialog_masked else ""
        a = " A" if self.is_accessory else ""
        return f"[Engine {self.tmcc_id} {self.unit_type.name.title()} {d}{hm}{dm}{tl}{a} (0b{bin(self.flags)})]"

    @property
    def info(self) -> str:
        d = "F" if self.is_forward else "R"
        tl = " T" if self.is_train_linked else ""
        hm = " H" if self.is_horn_masked else ""
        dm = " D" if self.is_dialog_masked else ""
        a = " A" if self.is_accessory else ""
        return f"{self.unit_type.name.title()[0]} {d}{hm}{dm}{tl}{a} {self.flags}"

    @property
    def unit_type(self) -> UnitBits:
        return UnitBits(self.flags & 0b11)

    @property
    def is_single(self) -> bool:
        return 0b11 & self.flags == 0b0

    @property
    def is_head(self) -> bool:
        return 0b11 & self.flags == 0b1

    @property
    def is_middle(self) -> bool:
        return 0b11 & self.flags == 0b10

    @property
    def is_tail(self) -> bool:
        return 0b11 & self.flags == 0b11

    @property
    def is_forward(self) -> bool:
        return 0b100 & self.flags == 0b000

    @property
    def is_reverse(self) -> bool:
        return 0b100 & self.flags == 0b100

    @property
    def is_train_linked(self) -> bool:
        return 0b1000 & self.flags == 0b1000

    @property
    def is_horn_masked(self) -> bool:
        return 0b10000 & self.flags == 0b10000

    @property
    def is_dialog_masked(self) -> bool:
        return 0b100000 & self.flags == 0b100000

    @property
    def is_tmcc2(self) -> bool:
        return 0b1000000 & self.flags == 0b1000000

    @property
    def is_accessory(self) -> bool:
        return 0b10000000 & self.flags == 0b10000000

    @property
    def as_bytes(self) -> bytes:
        byte_str = self.flags.to_bytes(1, byteorder="little")
        byte_str += self.tmcc_id.to_bytes(1, byteorder="little")
        return byte_str


class RouteComponent:
    @classmethod
    def from_bytes(cls, data: bytes) -> list[RouteComponent]:
        route_comps: list[RouteComponent] = list()
        data_len = len(data)
        for i in range(0, 32, 2):
            if data_len > i:
                if data[i] != 0xFF and data[i + 1] != 0xFF:
                    route_comps.insert(0, RouteComponent(tmcc_id=data[i + 1], flags=data[i]))
            else:
                break
        route_comps = sorted(route_comps, key=lambda s: s.tmcc_id)
        return route_comps

    @classmethod
    def to_bytes(cls, components: list[RouteComponent]) -> bytes:
        byte_str = bytes()
        for comp in components:
            byte_str += comp.as_bytes
        byte_str += b"\xff" * (32 - len(byte_str))
        return byte_str

    def __init__(self, tmcc_id: int, flags: int) -> None:
        self.tmcc_id = tmcc_id
        self.flags = flags

    @property
    def is_thru(self) -> bool:
        return 0x03 & self.flags == 0

    @property
    def is_out(self) -> bool:
        return self.is_thru is False

    @property
    def as_signature(self) -> dict[int, bool]:
        return {self.tmcc_id: self.is_thru}

    @property
    def as_request(self) -> CommandReq:
        cmd_enum = TMCC1SwitchCommandEnum.THRU if self.is_thru is True else TMCC1SwitchCommandEnum.OUT
        return CommandReq(cmd_enum, self.tmcc_id)

    @property
    def as_bytes(self) -> bytes:
        byte_str = self.flags.to_bytes(1, byteorder="little")
        byte_str += self.tmcc_id.to_bytes(1, byteorder="little")
        return byte_str
